"""Models and functions used for the LogsWidget."""
import re
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from html import escape as html_escape
from string import Template
from typing import Any, Dict, Optional, OrderedDict

from PySide2.QtCore import QPoint, QUrl, Qt, Signal
from PySide2.QtGui import QDesktopServices, QMouseEvent
from PySide2.QtWidgets import (
    QMenu,
    QPlainTextEdit,
    QVBoxLayout,
    QWidget,
)

from ywh2bt.gui.widgets.typing import as_signal_instance


class LogType(Enum):
    """Log type."""

    standard = 'normal'
    success = 'success'
    warning = 'warning'
    error = 'error'


LEVEL_COLORS: Dict[LogType, str] = OrderedDict[LogType, str]({
    LogType.standard: 'black',
    LogType.success: 'green',
    LogType.warning: 'orange',
    LogType.error: 'red',
})


@dataclass
class LogEntry:
    """A log entry."""

    date_time: datetime
    context: str
    message: str
    log_type: LogType = LogType.standard


class _PlainTextEdit(QPlainTextEdit):
    link_clicked: Signal = Signal(str)
    _clicked_anchor: Optional[str]

    def mousePressEvent(  # noqa: N802
        self,
        event: QMouseEvent,
    ) -> None:
        anchor = None
        if event.button() & Qt.LeftButton:
            anchor = self.anchorAt(
                event.pos(),
            )
        self._clicked_anchor = anchor
        super().mousePressEvent(
            event,
        )

    def mouseReleaseEvent(  # noqa: N802
        self,
        event: QMouseEvent,
    ) -> None:
        if event.button() & Qt.LeftButton:
            anchor = self.anchorAt(
                event.pos(),
            )
            if anchor and anchor == self._clicked_anchor:
                as_signal_instance(self.link_clicked).emit(
                    anchor,
                )
        super().mouseReleaseEvent(
            event,
        )


class LogsWidget(QWidget):
    """A widget to display log messages."""

    _text_edit: _PlainTextEdit

    def __init__(
        self,
        parent: Optional[QWidget] = None,
        *args: Any,
        **kwargs: Any,
    ):
        """
        Initialize self.

        Args:
            parent: a parent widget
            args: extra arguments
            kwargs: extra keyword arguments
        """
        super().__init__(
            parent,
            *args,
            *kwargs,
        )
        self._init_ui()

    def _init_ui(
        self,
    ) -> None:
        layout = QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)

        self._text_edit = self._create_text_edit()
        layout.addWidget(self._text_edit)

    def _create_text_edit(
        self,
    ) -> QPlainTextEdit:
        widget = _PlainTextEdit(self)
        widget.setLineWrapMode(QPlainTextEdit.NoWrap)
        widget.setReadOnly(True)
        widget.setContextMenuPolicy(Qt.CustomContextMenu)
        as_signal_instance(widget.customContextMenuRequested).connect(
            self._on_context_menu_requested,
        )
        as_signal_instance(widget.link_clicked).connect(
            self._on_link_clicked,
        )
        return widget

    def _on_context_menu_requested(
        self,
        position: QPoint,
    ) -> None:
        menu = QMenu(self)
        action_clear = menu.addAction('Clear')
        as_signal_instance(action_clear.triggered).connect(
            self._on_clear_action_triggered,
        )
        menu.popup(
            self.mapToGlobal(position),
        )

    def _on_clear_action_triggered(
        self,
    ) -> None:
        self._text_edit.setPlainText('')

    def _on_link_clicked(
        self,
        link: str,
    ) -> None:
        url = QUrl(link)
        QDesktopServices.openUrl(url)

    def add_log_entry(
        self,
        entry: LogEntry,
    ) -> None:
        """
        Add a new entry to the log.

        Args:
            entry: an entry
        """
        self._add_log_entry_html(
            entry=entry,
        )

    def _add_log_entry_html(
        self,
        entry: LogEntry,
    ) -> None:
        new_log_html = _format_log_entry_html(
            entry=entry,
        )
        self._text_edit.appendHtml(
            new_log_html,
        )


log_html_template_contents = """
<p style="color: ${type_color}"><b>[${date_time}]</b>: ${context}</p>
<p style="color: ${type_color}; white-space: pre">${message}</p>
<p></p>
"""
log_html_template = Template(log_html_template_contents)
log_html_link_template = Template('<a href="${url}">${url}</a>')

url_re = re.compile(r'(https?://[^\s]+)')


def _format_log_entry_html(
    entry: LogEntry,
) -> str:
    date_time = entry.date_time.strftime('%Y-%m-%d %H:%M:%S')  # noqa: WPS323
    context = html_escape(entry.context.strip())
    message = html_escape(entry.message.strip())
    urls = url_re.findall(message)
    for url in set(urls):
        link = log_html_link_template.substitute(
            url=url,
        )
        message = message.replace(
            url,
            link,
        )
    return log_html_template.substitute(
        date_time=date_time,
        context=context,
        message=message,
        type_color=_log_type_to_color(
            log_type=entry.log_type,
        ),
    )


def _log_type_to_color(
    log_type: LogType,
) -> str:
    return LEVEL_COLORS.get(
        log_type,
        'black',
    )



"""Models and functions used for testing of YesWeHack and tracker clients."""
from typing import (
    Optional,
    cast,
)

from ywh2bt.core.api.tracker import TrackerClientError
from ywh2bt.core.api.yeswehack import YesWeHackApiClientError
from ywh2bt.core.configuration.root import RootConfiguration
from ywh2bt.core.configuration.tracker import (
    TrackerConfiguration,
    Trackers,
)
from ywh2bt.core.configuration.yeswehack import (
    YesWeHackConfiguration,
    YesWeHackConfigurations,
)
from ywh2bt.core.factories.tracker_clients import (
    TrackerClientsAbstractFactory,
)
from ywh2bt.core.factories.yeswehack_api_clients import (
    YesWeHackApiClientsAbstractFactory,
)
from ywh2bt.core.tester.error import TesterError
from ywh2bt.core.tester.listener import (
    NoOpTesterListener,
    TesterEndEvent,
    TesterEndTrackerEvent,
    TesterEndYesWeHackEvent,
    TesterListener,
    TesterStartEvent,
    TesterStartTrackerEvent,
    TesterStartYesWeHackEvent,
)


class Tester:
    """A class used for testing of YesWeHack and tracker clients."""

    _configuration: RootConfiguration
    _yes_we_hack_api_clients_factory: YesWeHackApiClientsAbstractFactory
    _tracker_clients_factory: TrackerClientsAbstractFactory
    _listener: TesterListener

    def __init__(
        self,
        configuration: RootConfiguration,
        yes_we_hack_api_clients_factory: YesWeHackApiClientsAbstractFactory,
        tracker_clients_factory: TrackerClientsAbstractFactory,
        listener: Optional[TesterListener] = None,
    ):
        """
        Initialize self.

        Args:
            configuration: a configuration
            yes_we_hack_api_clients_factory: a YesWeHackApiClients factory
            tracker_clients_factory: a TrackerClients factory
            listener: an observer that will receive test events
        """
        self._configuration = configuration
        self._yes_we_hack_api_clients_factory = yes_we_hack_api_clients_factory
        self._tracker_clients_factory = tracker_clients_factory
        self._listener = listener or NoOpTesterListener()

    def test(
        self,
    ) -> None:
        """Test YesWeHack and tracker clients."""
        self._listener.on_event(
            event=TesterStartEvent(
                configuration=self._configuration,
            ),
        )
        self._test_yeswehack()
        self._test_bugtrackers()
        self._listener.on_event(
            event=TesterEndEvent(
                configuration=self._configuration,
            ),
        )

    def _test_yeswehack(
        self,
    ) -> None:
        yeswehack_configurations = cast(YesWeHackConfigurations, self._configuration.yeswehack)
        for yeswehack_name, yeswehack_configuration in yeswehack_configurations.items():
            self._test_yeswehack_configuration(
                yeswehack_name=yeswehack_name,
                yeswehack_configuration=yeswehack_configuration,
            )

    def _test_yeswehack_configuration(
        self,
        yeswehack_name: str,
        yeswehack_configuration: YesWeHackConfiguration,
    ) -> None:
        self._listener.on_event(
            event=TesterStartYesWeHackEvent(
                yeswehack_name=yeswehack_name,
                yeswehack_configuration=yeswehack_configuration,
            ),
        )
        yeswehack_client = self._yes_we_hack_api_clients_factory.get_yeswehack_api_client(
            configuration=yeswehack_configuration,
        )
        try:
            yeswehack_client.test()
        except YesWeHackApiClientError as e:
            raise TesterError(f'Test for yeswehack {yeswehack_name} has failed') from e
        self._listener.on_event(
            event=TesterEndYesWeHackEvent(
                yeswehack_name=yeswehack_name,
                yeswehack_configuration=yeswehack_configuration,
            ),
        )

    def _test_bugtrackers(
        self,
    ) -> None:
        bugtracker_configurations = cast(Trackers, self._configuration.bugtrackers)
        for bugtracker_name, bugtracker_configuration in bugtracker_configurations.items():
            self._test_bugtracker(
                bugtracker_name=bugtracker_name,
                bugtracker_configuration=bugtracker_configuration,
            )

    def _test_bugtracker(
        self,
        bugtracker_name: str,
        bugtracker_configuration: TrackerConfiguration,
    ) -> None:
        self._listener.on_event(
            event=TesterStartTrackerEvent(
                tracker_name=bugtracker_name,
                tracker_configuration=bugtracker_configuration,
            ),
        )
        tracker_client = self._tracker_clients_factory.get_tracker_client(
            configuration=bugtracker_configuration,
        )
        try:
            tracker_client.test()
        except TrackerClientError as e:
            raise TesterError(f'Test for bugtracker {bugtracker_name} has failed') from e
        self._listener.on_event(
            event=TesterEndTrackerEvent(
                tracker_name=bugtracker_name,
                tracker_configuration=bugtracker_configuration,
            ),
        )



"""Models and functions related to GUI tester thread."""
from datetime import datetime
from typing import cast

from PySide2.QtCore import (
    QObject,
    QThread,
    Signal,
    SignalInstance,
)
from singledispatchmethod import singledispatchmethod

from ywh2bt.core.configuration.root import RootConfiguration
from ywh2bt.core.error import error_to_string
from ywh2bt.core.exceptions import CoreException
from ywh2bt.core.factories.tracker_clients import TrackerClientsFactory
from ywh2bt.core.factories.yeswehack_api_clients import YesWeHackApiClientsFactory
from ywh2bt.core.tester.listener import (
    TesterEndEvent,
    TesterEndTrackerEvent,
    TesterEndYesWeHackEvent,
    TesterEvent,
    TesterListener,
    TesterStartEvent,
    TesterStartTrackerEvent,
    TesterStartYesWeHackEvent,
)
from ywh2bt.core.tester.tester import Tester
from ywh2bt.gui.widgets.logs_widget import (
    LogEntry,
    LogType,
)
from ywh2bt.gui.widgets.root_configuration_entry import RootConfigurationEntry
from ywh2bt.gui.widgets.typing import as_signal_instance


class TesterThread(QThread):
    """A thread for testing configuration."""

    test_started: Signal = Signal()
    test_ended: Signal = Signal()

    _entry: RootConfigurationEntry
    _log_entry_available_signal: SignalInstance
    _tester: Tester

    def __init__(
        self,
        parent: QObject,
        entry: RootConfigurationEntry,
        log_entry_available_signal: SignalInstance,
    ):
        """
        Initialize self.

        Args:
            parent: a parent widget
            entry: a root configuration entry
            log_entry_available_signal: a signal that receive events from the thread
        """
        super().__init__(
            parent,
        )
        self._entry = entry
        self._log_entry_available_signal = log_entry_available_signal
        observer = _TesterListener(
            entry=entry,
            log_entry_available_signal=log_entry_available_signal,
        )
        self._tester = Tester(
            configuration=cast(RootConfiguration, entry.configuration),
            yes_we_hack_api_clients_factory=YesWeHackApiClientsFactory(),
            tracker_clients_factory=TrackerClientsFactory(),
            listener=observer,
        )

    def run(
        self,
    ) -> None:
        """Run the thread."""
        as_signal_instance(self.test_started).emit()
        try:
            self._tester.test()
        except CoreException as e:
            self._log_entry_available_signal.emit(
                LogEntry(
                    date_time=datetime.now(),
                    context=f'{self._entry.name}',
                    message=error_to_string(
                        error=e,
                    ),
                    log_type=LogType.error,
                ),
            )
        as_signal_instance(self.test_ended).emit()


class _TesterListener(TesterListener):
    _entry: RootConfigurationEntry
    _log_entry_available_signal: SignalInstance

    def __init__(
        self,
        entry: RootConfigurationEntry,
        log_entry_available_signal: SignalInstance,
    ):
        self._entry = entry
        self._log_entry_available_signal = log_entry_available_signal

    def _log_message(
        self,
        message: str,
        log_type: LogType = LogType.standard,
    ) -> None:
        self._log_entry_available_signal.emit(
            LogEntry(
                date_time=datetime.now(),
                context=f'{self._entry.name}',
                message=message,
                log_type=log_type,
            ),
        )

    def on_event(
        self,
        event: TesterEvent,
    ) -> None:
        self._on_event(event)

    @singledispatchmethod
    def _on_event(
        self,
        event: TesterEvent,
    ) -> None:
        self._log_message(
            message=f'Unhandled event {event}',
        )

    @_on_event.register
    def _on_start(
        self,
        event: TesterStartEvent,
    ) -> None:
        self._log_message(
            message='Starting test',
        )

    @_on_event.register
    def _on_end(
        self,
        event: TesterEndEvent,
    ) -> None:
        self._log_message(
            log_type=LogType.success,
            message='Test done',
        )

    @_on_event.register
    def _on_start_test_yeswehack(
        self,
        event: TesterStartYesWeHackEvent,
    ) -> None:
        self._log_message(
            message=f'Testing YesWeHack {event.yeswehack_name}',
        )

    @_on_event.register
    def _on_end_test_yeswehack(
        self,
        event: TesterEndYesWeHackEvent,
    ) -> None:
        self._log_message(
            log_type=LogType.success,
            message=f'YesWeHack {event.yeswehack_name}: OK',
        )

    @_on_event.register
    def _on_start_test_tracker(
        self,
        event: TesterStartTrackerEvent,
    ) -> None:
        self._log_message(
            message=f'Testing tracker {event.tracker_name}',
        )

    @_on_event.register
    def _on_end_test_tracker(
        self,
        event: TesterEndTrackerEvent,
    ) -> None:
        self._log_message(
            log_type=LogType.success,
            message=f'Tracker {event.tracker_name}: OK',
        )



"""Models and functions used for data synchronisation between YesWeHack and Jira trackers."""
import re
from copy import deepcopy
from datetime import datetime
from string import Template
from typing import (
    Dict,
    List,
    Optional,
    Tuple,
    cast,
)
from urllib.parse import quote

from jira import (
    Comment as JIRAComment,
    Issue as JIRAIssue,
    JIRA,
    JIRAError,
)
from requests_toolbelt.multipart.encoder import CustomBytesIO  # type: ignore

from ywh2bt.core.api.formatter.markdown import ReportMessageMarkdownFormatter
from ywh2bt.core.api.models.report import (
    Attachment,
    Log,
    Report,
)
from ywh2bt.core.api.tracker import (
    SendLogsResult,
    TrackerAttachment,
    TrackerClient,
    TrackerClientError,
    TrackerIssue,
    TrackerIssueComment,
    TrackerIssueComments,
)
from ywh2bt.core.api.trackers.jira.formatter import JiraReportMessageFormatter
from ywh2bt.core.configuration.trackers.jira import JiraConfiguration
from ywh2bt.core.converter.jira2markdown import jira2markdown

_RE_IMAGE = re.compile(pattern=r'!([^!|]+)(?:\|[^!]*)?!')
_TITLE_MAX_SIZE = 255
_TEXT_MAX_SIZE = 32767


class JiraTrackerClientError(TrackerClientError):
    """A Jira tracker client error."""


class JiraTrackerClient(TrackerClient[JiraConfiguration]):
    """A Jira tracker client."""

    _jira: Optional[JIRA]
    _message_formatter: JiraReportMessageFormatter

    _attachments_list_description_item_jira_template = Template('- [${name}|${url}]')
    _attachments_list_description_item_markdown_template = Template('- [${name}](${url})')

    def __init__(
        self,
        configuration: JiraConfiguration,
    ):
        """
        Initialize self.

        Args:
            configuration: a Jira configuration
        """
        super().__init__(
            configuration=configuration,
        )
        self._jira = None
        self._message_formatter = JiraReportMessageFormatter()

    @property
    def tracker_type(self) -> str:
        """
        Get the type of the tracker client.

        Returns:
            the type of the  tracker client
        """
        return 'Jira'

    def _build_tracker_issue(
        self,
        issue_id: str,
        issue_url: str,
        closed: bool,
    ) -> TrackerIssue:
        return TrackerIssue(
            tracker_url=cast(str, self.configuration.url),
            project=cast(str, self.configuration.project),
            issue_id=issue_id,
            issue_url=issue_url,
            closed=closed,
        )

    def get_tracker_issue(
        self,
        issue_id: str,
    ) -> Optional[TrackerIssue]:
        """
        Get a tracker issue.

        Args:
            issue_id: an issue id

        Returns:
            The issue if it exists, else None
        """
        try:
            jira_issue = self._get_issue(
                issue_id=issue_id,
            )
        except JiraTrackerClientError:
            return None
        return self._build_tracker_issue(
            issue_id=issue_id,
            issue_url=jira_issue.permalink(),
            closed=self._issue_is_closed(
                jira_issue=jira_issue,
            ),
        )

    def _issue_is_closed(
        self,
        jira_issue: JIRAIssue,
    ) -> bool:
        fields = jira_issue.fields
        return fields.status is not None and fields.status.name == self.configuration.issue_closed_status

    def get_tracker_issue_comments(
        self,
        issue_id: str,
        exclude_comments: Optional[List[str]] = None,
    ) -> TrackerIssueComments:
        """
        Get a list of comments on an issue.

        Args:
            issue_id: an issue id
            exclude_comments: an optional list of comment to exclude

        Returns:
            The list of comments
        """
        try:
            jira_issue = self._get_issue(
                issue_id=issue_id,
            )
        except JiraTrackerClientError:
            return []
        return self._extract_comments(
            jira_issue=jira_issue,
            exclude_comments=exclude_comments,
        )

    def _extract_comments(
        self,
        jira_issue: JIRAIssue,
        exclude_comments: Optional[List[str]] = None,
    ) -> List[TrackerIssueComment]:
        return [
            self._extract_comment(
                jira_issue=jira_issue,
                jira_comment=jira_comment,
            )
            for jira_comment in jira_issue.fields.comment.comments
            if exclude_comments is None or str(jira_comment.id) not in exclude_comments
        ]

    def _extract_comment(
        self,
        jira_issue: JIRAIssue,
        jira_comment: JIRAComment,
    ) -> TrackerIssueComment:
        jira_body = jira_comment.body
        inline_images = _RE_IMAGE.findall(jira_body)
        comment_attachments: Dict[str, TrackerAttachment] = {}
        for jira_attachment in jira_issue.fields.attachment:
            if jira_attachment.filename in inline_images:
                comment_attachments[jira_attachment.filename] = TrackerAttachment(
                    filename=jira_attachment.filename,
                    mime_type=jira_attachment.mimeType,
                    content=jira_attachment.get(),
                )
        return TrackerIssueComment(
            created_at=self._parse_date(
                date=jira_comment.created,
            ),
            author=jira_comment.author.displayName,
            comment_id=str(jira_comment.id),
            body=jira2markdown(
                src=jira_body,
            ),
            attachments=comment_attachments,
        )

    def send_report(
        self,
        report: Report,
    ) -> TrackerIssue:
        """
        Send a report to the tracker.

        Args:
            report: a report

        Returns:
            information about the sent report
        """
        self._ensure_auth()
        title = self._message_formatter.format_report_title(
            report=report,
        )
        if len(title) > _TITLE_MAX_SIZE:
            title = f'{title[:_TITLE_MAX_SIZE - 3]}...'
        description = self._message_formatter.format_report_description(
            report=report,
        ) + self._get_attachments_list_description(
            title='*Attachments*:',
            item_template=self._attachments_list_description_item_jira_template,
            attachments=report.attachments,
        )
        markdown_description = ''
        description_attachment = None
        if len(description) > _TEXT_MAX_SIZE:
            description_attachment = self._build_external_description_attachment(
                name=f'report-{report.local_id.replace("#", "")}-description.md',
            )
            markdown_description = ReportMessageMarkdownFormatter().format_report_description(
                report=report,
            ) + self._get_attachments_list_description(
                title='**Attachments**:',
                item_template=self._attachments_list_description_item_markdown_template,
                attachments=report.attachments,
            )
            report_copy = deepcopy(report)
            report_copy.description_html = (
                '<p>This report description is too large to fit into a JIRA issue. '
                + f'See attachment <a href="{description_attachment.url}">{description_attachment.original_name}</a> '
                + 'for more details.</p>'
            )
            description = self._message_formatter.format_report_description(
                report=report_copy,
            ) + self._get_attachments_list_description(
                title='*Attachments*:',
                item_template=self._attachments_list_description_item_jira_template,
                attachments=[
                    description_attachment,
                    *report.attachments,
                ],
            )
        jira_issue = self._create_issue(
            title=title,
            description='This issue is being synchronized. Please check back in a moment.',
        )
        description, markdown_description = self._replace_attachments_references(
            uploads=self._upload_attachments(
                issue=jira_issue,
                attachments=report.attachments,
            ),
            referencing_texts=[
                description,
                markdown_description,
            ],
        )
        if description_attachment:
            description_attachment.data_loader = lambda: bytes(markdown_description, 'utf-8')
            description = self._replace_attachments_references(
                uploads=self._upload_attachments(
                    issue=jira_issue,
                    attachments=[
                        description_attachment,
                    ],
                ),
                referencing_texts=[
                    description,
                ],
            )[0]
        jira_issue.update(
            description=description,
        )
        return self._build_tracker_issue(
            issue_id=jira_issue.key,
            issue_url=jira_issue.permalink(),
            closed=False,
        )

    def _build_external_description_attachment(
        self,
        name: str,
    ) -> Attachment:
        return Attachment(
            attachment_id=0,
            name=name,
            original_name=name,
            mime_type='text/markdown',
            size=0,
            url=f'http://tracker/external/{name}',
            data_loader=lambda: bytes('', 'utf-8'),
        )

    def send_logs(
        self,
        tracker_issue: TrackerIssue,
        logs: List[Log],
    ) -> SendLogsResult:
        """
        Send logs to the tracker.

        Args:
            tracker_issue: information about the tracker issue
            logs: a list of logs

        Returns:
            information about the sent logs
        """
        jira_issue = self._get_issue(
            issue_id=tracker_issue.issue_id,
        )
        tracker_comments = SendLogsResult(
            tracker_issue=tracker_issue,
            added_comments=[],
        )
        for log in logs:
            jira_comment = self._add_comment(
                issue=jira_issue,
                log=log,
            )
            tracker_comments.added_comments.append(
                TrackerIssueComment(
                    created_at=self._parse_date(
                        date=jira_comment.created,
                    ),
                    author=jira_comment.author.displayName,
                    comment_id=str(jira_comment.id),
                    body=jira2markdown(
                        src=jira_comment.body,
                    ),
                    attachments={},
                ),
            )
        return tracker_comments

    def _parse_date(
        self,
        date: str,
    ) -> datetime:
        return datetime.strptime(
            date,
            '%Y-%m-%dT%H:%M:%S.%f%z',
        )

    def test(
        self,
    ) -> None:
        """Test the client."""
        self._ensure_auth()

    def _get_issue(
        self,
        issue_id: str,
    ) -> JIRAIssue:
        try:
            return self._get_client().issue(
                id=issue_id,
            )
        except JIRAError as e:
            raise JiraTrackerClientError(
                f'Unable to get JIRA issue {issue_id} in project {self.configuration.project}',
            ) from e

    def _create_issue(
        self,
        title: str,
        description: str,
    ) -> JIRAIssue:
        fields = {
            'project': {
                'key': self.configuration.project,
            },
            'summary': title,
            'description': description,
            'issuetype': {
                'name': self.configuration.issuetype,
            },
        }
        try:
            return self._get_client().create_issue(
                fields=fields,
            )
        except JIRAError as e:
            raise JiraTrackerClientError(
                f'Unable to create JIRA issue for project {self.configuration.project}',
            ) from e

    def _add_comment(
        self,
        issue: JIRAIssue,
        log: Log,
    ) -> JIRAComment:
        comment_body = self._message_formatter.format_log(
            log=log,
        ) + self._get_attachments_list_description(
            title='*Attachments*:',
            item_template=self._attachments_list_description_item_jira_template,
            attachments=log.attachments,
        )
        markdown_description = ''
        body_attachment = None
        if len(comment_body) > _TEXT_MAX_SIZE:
            body_attachment = self._build_external_description_attachment(
                name=f'comment-{log.log_id}-description.md',
            )
            markdown_description = ReportMessageMarkdownFormatter().format_log(
                log=log,
            ) + self._get_attachments_list_description(
                title='**Attachments**:',
                item_template=self._attachments_list_description_item_markdown_template,
                attachments=log.attachments,
            )
            log_copy = deepcopy(log)
            log_copy.message_html = (
                '<p>This comment is too large to fit into a JIRA comment. '
                + f'See attachment <a href="{body_attachment.url}">{body_attachment.original_name}</a> '
                + 'for more details.</p>'
            )
            comment_body = self._message_formatter.format_log(
                log=log_copy,
            ) + self._get_attachments_list_description(
                title='*Attachments*:',
                item_template=self._attachments_list_description_item_jira_template,
                attachments=[
                    body_attachment,
                    *log.attachments,
                ],
            )
        comment_body, markdown_description = self._replace_attachments_references(
            uploads=self._upload_attachments(
                issue=issue,
                attachments=log.attachments,
            ),
            referencing_texts=[
                comment_body,
                markdown_description,
            ],
        )
        if body_attachment:
            body_attachment.data_loader = lambda: bytes(markdown_description, 'utf-8')
            comment_body = self._replace_attachments_references(
                uploads=self._upload_attachments(
                    issue=issue,
                    attachments=[
                        body_attachment,
                    ],
                ),
                referencing_texts=[
                    comment_body,
                ],
            )[0]
        try:
            return self._get_client().add_comment(
                issue=str(issue),
                body=comment_body,
            )
        except JIRAError as e:
            raise JiraTrackerClientError(
                f'Unable to add JIRA comment for issue {issue} in project {self.configuration.project}',
            ) from e

    def _upload_attachments_and_substitute_references(
        self,
        issue: JIRAIssue,
        attachments: List[Attachment],
        referencing_texts: List[str],
    ) -> List[str]:
        uploads = self._upload_attachments(
            issue=issue,
            attachments=attachments,
        )
        for attachment, upload_url in uploads:
            referencing_texts = [
                text.replace(
                    attachment.url,
                    upload_url,
                )
                for text in referencing_texts
            ]
        return referencing_texts

    def _replace_attachments_references(
        self,
        uploads: List[Tuple[Attachment, str]],
        referencing_texts: List[str],
    ) -> List[str]:
        for attachment, upload_url in uploads:
            referencing_texts = [
                text.replace(
                    attachment.url,
                    upload_url,
                )
                for text in referencing_texts
            ]
        return referencing_texts

    def _upload_attachments(
        self,
        issue: JIRAIssue,
        attachments: List[Attachment],
    ) -> List[Tuple[Attachment, str]]:
        uploads = []
        for attachment in attachments:
            try:
                issue_attachment = self._get_client().add_attachment(
                    issue=issue.key,
                    filename=attachment.original_name,
                    attachment=CustomBytesIO(buffer=attachment.data),  # using CustomBytesIO despite jira expectations
                )
            except JIRAError as e:
                raise JiraTrackerClientError(
                    f'Unable to upload attachments for project {self.configuration.project} to JIRA',
                ) from e
            issue_attachment_url = quote(
                issue_attachment.content,
                safe=':/?&=',
            )
            uploads.append(
                (
                    attachment,
                    issue_attachment_url,
                ),
            )
        return uploads

    def _get_attachments_list_description(
        self,
        title: str,
        item_template: Template,
        attachments: List[Attachment],
    ) -> str:
        attachments_lines = []
        if attachments:
            attachments_lines = [
                '',
                title,
            ]
            for attachment in attachments:
                attachments_lines.append(
                    item_template.substitute(
                        name=attachment.original_name,
                        url=attachment.url,
                    ),
                )
            attachments_lines.append('')
        return '\n'.join(attachments_lines)

    def _get_client(
        self,
    ) -> JIRA:
        if not self._jira:
            self._jira = self._build_client()
        return self._jira

    def _build_client(
        self,
    ) -> JIRA:
        configuration = self.configuration
        try:
            return JIRA(
                server=cast(str, configuration.url),
                basic_auth=(
                    cast(str, configuration.login),
                    cast(str, configuration.password),
                ),
                options={
                    'verify': configuration.verify,
                },
            )
        except JIRAError as e:
            raise JiraTrackerClientError('Unable to connect to JIRA') from e

    def _ensure_auth(
        self,
    ) -> None:
        try:
            self._get_client().myself()
        except JIRAError as e:
            raise JiraTrackerClientError('Unable to authenticate to JIRA') from e



"""Models and functions related to GUI synchronizer thread."""
from datetime import datetime
from typing import cast

from PySide2.QtCore import (
    QObject,
    QThread,
    Signal,
    SignalInstance,
)
from singledispatchmethod import singledispatchmethod

from ywh2bt.core.api.models.report import REPORT_STATUS_TRANSLATIONS
from ywh2bt.core.configuration.root import RootConfiguration
from ywh2bt.core.error import error_to_string
from ywh2bt.core.exceptions import CoreException
from ywh2bt.core.factories.tracker_clients import TrackerClientsFactory
from ywh2bt.core.factories.yeswehack_api_clients import YesWeHackApiClientsFactory
from ywh2bt.core.synchronizer.listener import (
    SynchronizerEndEvent,
    SynchronizerEndFetchReportsEvent,
    SynchronizerEndSendReportEvent,
    SynchronizerEvent,
    SynchronizerListener,
    SynchronizerStartEvent,
    SynchronizerStartFetchReportsEvent,
    SynchronizerStartSendReportEvent,
)
from ywh2bt.core.synchronizer.synchronizer import Synchronizer
from ywh2bt.gui.widgets.logs_widget import (
    LogEntry,
    LogType,
)
from ywh2bt.gui.widgets.root_configuration_entry import RootConfigurationEntry
from ywh2bt.gui.widgets.typing import as_signal_instance


class SynchronizerThread(QThread):
    """A thread for synchronization."""

    synchronization_started: Signal = Signal()
    synchronization_ended: Signal = Signal()

    _entry: RootConfigurationEntry
    _log_entry_available_signal: SignalInstance
    _synchronizer: Synchronizer

    def __init__(
        self,
        parent: QObject,
        entry: RootConfigurationEntry,
        log_entry_available_signal: SignalInstance,
    ):
        """
        Initialize self.

        Args:
            parent: a parent widget
            entry: a root configuration entry
            log_entry_available_signal: a signal that receive events from the thread
        """
        super().__init__(
            parent,
        )
        self._entry = entry
        self._log_entry_available_signal = log_entry_available_signal
        observer = _SynchronizerListener(
            entry=entry,
            log_entry_available_signal=log_entry_available_signal,
        )
        self._synchronizer = Synchronizer(
            configuration=cast(RootConfiguration, entry.configuration),
            yes_we_hack_api_clients_factory=YesWeHackApiClientsFactory(),
            tracker_clients_factory=TrackerClientsFactory(),
            listener=observer,
        )

    def run(
        self,
    ) -> None:
        """Run the thread."""
        as_signal_instance(self.synchronization_started).emit()
        try:
            self._synchronizer.synchronize()
        except CoreException as e:
            self._log_entry_available_signal.emit(
                LogEntry(
                    date_time=datetime.now(),
                    context=f'{self._entry.name}',
                    message=error_to_string(
                        error=e,
                    ),
                    log_type=LogType.error,
                ),
            )
        as_signal_instance(self.synchronization_ended).emit()


class _SynchronizerListener(SynchronizerListener):
    _entry: RootConfigurationEntry
    _log_entry_available_signal: SignalInstance

    def __init__(
        self,
        entry: RootConfigurationEntry,
        log_entry_available_signal: SignalInstance,
    ):
        self._entry = entry
        self._log_entry_available_signal = log_entry_available_signal

    def _log_message(
        self,
        message: str,
        log_type: LogType = LogType.standard,
    ) -> None:
        self._log_entry_available_signal.emit(
            LogEntry(
                date_time=datetime.now(),
                context=f'{self._entry.name}',
                message=message,
                log_type=log_type,
            ),
        )

    def on_event(
        self,
        event: SynchronizerEvent,
    ) -> None:
        self._on_event(event)

    @singledispatchmethod
    def _on_event(
        self,
        event: SynchronizerEvent,
    ) -> None:
        self._log_message(
            message=f'Unhandled event {event}',
        )

    @_on_event.register
    def _on_start(
        self,
        event: SynchronizerStartEvent,
    ) -> None:
        self._log_message(
            message='Starting synchronization...',
        )

    @_on_event.register
    def _on_end(
        self,
        event: SynchronizerEndEvent,
    ) -> None:
        self._log_message(
            log_type=LogType.success,
            message='Synchronization done.',
        )

    @_on_event.register
    def _on_start_fetch_reports(
        self,
        event: SynchronizerStartFetchReportsEvent,
    ) -> None:
        yeswehack_name = event.yeswehack_name
        program_slug = event.program_slug
        self._log_message(
            message=f'Processing YesWeHack "{yeswehack_name}"...',
        )
        self._log_message(
            message=f'Fetching reports for program "{program_slug}"...',
        )

    @_on_event.register
    def _on_end_fetch_reports(
        self,
        event: SynchronizerEndFetchReportsEvent,
    ) -> None:
        reports_count = len(event.reports)
        self._log_message(
            log_type=LogType.success,
            message=f'{reports_count} report(s) fetched.',
        )

    @_on_event.register
    def _on_start_send_report(
        self,
        event: SynchronizerStartSendReportEvent,
    ) -> None:
        report = event.report
        if report.title:
            report_details = f'#{report.report_id} ({report.title})'
        else:
            report_details = f'#{report.report_id}'
        self._log_message(
            message=f'Processing report {report_details} with "{event.tracker_name}"...',
        )

    @_on_event.register
    def _on_end_send_report(
        self,
        event: SynchronizerEndSendReportEvent,
    ) -> None:
        tracker_issue = event.tracker_issue
        issue_added_comment_count = len(event.issue_added_comments)
        if not tracker_issue or not event.is_existing_issue:
            issue_details = [
                'does not exist anymore',
            ]
        else:
            issue_details = [
                tracker_issue.issue_url,
            ]
            if event.is_created_issue:
                if issue_added_comment_count:
                    issue_details.append('updated')
            else:
                issue_details.append('added')
            if issue_added_comment_count:
                issue_details.append(f'{issue_added_comment_count} comment(s) added')
        report_added_comment_count = len(event.report_added_comments)
        report_details = []
        if report_added_comment_count:
            report_details.append(f'{report_added_comment_count} comment(s) added')
        report_details.append(f'tracking status {"updated" if event.tracking_status_updated else "unchanged"}')
        if event.new_report_status:
            old_status, new_status = event.new_report_status
            old_status_translation = REPORT_STATUS_TRANSLATIONS.get(old_status, 'Unknown')
            new_status_translation = REPORT_STATUS_TRANSLATIONS.get(new_status, 'Unknown')
            report_details.append(f'status "{old_status_translation}" -> "{new_status_translation}"')
        message = ' | '.join(
            (
                f'issue => {" ; ".join(issue_details)}',
                f'report => {" ; ".join(report_details)}',
            ),
        )
        if event.report.title:
            report_description = f'#{event.report.report_id} ({event.report.title})'
        else:
            report_description = f'#{event.report.report_id}'
        self._log_message(
            log_type=LogType.success,
            message=f'Processed report {report_description} with "{event.tracker_name}": {message}.',
        )



from typing import (
    List,
    Union,
)


def _sizeof_fmt(
    num: Union[int, float],
    base: int,
    units: List[str],
    max_unit: str,
    precision: int = 1,
    suffix: str = 'B'
) -> str:
    for unit in units:
        if abs(num) < base:
            return f'{num:.{precision}f}{unit}{suffix}'
        num /= base
    return f'{num:.{precision}f}{max_unit}{suffix}'


def sizeof_fmt_iec(num: Union[int, float], precision: int = 1, suffix: str = 'B') -> str:
    return _sizeof_fmt(
        num,
        base=1024,
        units=['', 'Ki', 'Mi', 'Gi', 'Ti', 'Pi', 'Ei', 'Zi'],
        max_unit='Yi',
        precision=precision,
        suffix=suffix,
    )


def sizeof_fmt_si(num: Union[int, float], precision: int = 1, suffix: str = 'B') -> str:
    return _sizeof_fmt(
        num,
        base=1000,
        units=['', 'k', 'M', 'G', 'T', 'P', 'E', 'Z'],
        max_unit='Y',
        precision=precision,
        suffix=suffix,
    )



"""Models and functions used in file dialogs."""
from string import Template
from typing import Dict, Optional, Set

from ywh2bt.core.core import AvailableFormat


class FileFormatDialogFilters:
    """A class that holds data for manipulating file dialogs filters."""

    _filters: Dict[str, str]

    def __init__(
        self,
        formats: Dict[str, AvailableFormat],
    ):
        """
        Initialize self.

        Args:
            formats: a dict of available core formats
        """
        filter_template = Template('${name} files (${patterns})')
        filters = {}
        for name, available_format in formats.items():
            patterns = self._get_extensions_patterns(
                extensions=available_format.extensions,
            )
            filters[name] = filter_template.substitute(
                name=name.upper(),
                patterns=patterns,
            )
        self._filters = filters

    @classmethod
    def _get_extensions_patterns(
        cls,
        extensions: Set[str],
    ) -> str:
        template = Template('*.${extension}')
        patterns = []
        for extension in extensions:
            patterns.append(
                template.substitute(
                    extension=extension,
                ),
            )
        return ' '.join(patterns)

    def get_filters_string_for(
        self,
        format_name: str,
    ) -> Optional[str]:
        """
        Get the filters for a format.

        Args:
            format_name: a format name

        Returns:
            The filters
        """
        return self._filters.get(format_name)

    def get_filters_string(
        self,
        separator: str = ';;',
    ) -> str:
        """
        Get the filters.

        Args:
            separator: a string to be used as a separators between the filters

        Returns:
            The filters
        """
        return separator.join(self._filters.values())

    def get_format_name(
        self,
        filter_string: str,
    ) -> Optional[str]:
        """
        Get the name of format given a filter.

        Args:
            filter_string: a filter string

        Returns:
            The format name or None if the format does not exist
        """
        for format_name, existing_filter_string in self._filters.items():
            if filter_string == existing_filter_string:
                return format_name
        return None



"""Models and functions used for data synchronisation between YesWeHack and ServiceNow trackers."""
import asyncio
import datetime
import re
from abc import ABC
from copy import deepcopy
from dataclasses import dataclass
from string import Template
from typing import (
    Any,
    Iterable,
    List,
    Optional,
    Set,
    Tuple,
    Union,
    cast,
)

from aiosnow import (
    Client,
    select,
)
from aiosnow.exceptions import AiosnowException
from aiosnow.models.table.declared import (
    IncidentModel,
    JournalModel,
)

from ywh2bt.core.api.formatter.markdown import ReportMessageMarkdownFormatter
from ywh2bt.core.api.models.report import (
    Attachment,
    Log,
    Report,
)
from ywh2bt.core.api.tracker import (
    SendLogsResult,
    TrackerAttachment,
    TrackerClient,
    TrackerClientError,
    TrackerIssue,
    TrackerIssueComment,
    TrackerIssueComments,
)
from ywh2bt.core.api.trackers.servicenow.formatter import ServiceNowReportMessageFormatter
from ywh2bt.core.api.trackers.servicenow.model import (
    InMemoryAttachmentModel,
    Record,
    UserModel,
)
from ywh2bt.core.configuration.trackers.servicenow import ServiceNowConfiguration

_RE_INLINE_ATTACHMENT = re.compile(pattern=r'(!?\[([^\]]+)]\(([^)]+)\))')
_TEXT_MAX_SIZE = 32767  # ServiceNow has no defined limit but API sometime times out with big contents

_MODEL_NOT_EQUALS_LIMIT = 100


@dataclass
class RecordWrapper(ABC):
    """A wrapper for records."""

    data: Record


@dataclass
class CommentRecordWrapper(RecordWrapper):
    """A wrapper for comment records."""


@dataclass
class AttachmentRecordWrapper(RecordWrapper):
    """A wrapper for attachment records."""


class ServiceNowTrackerClientError(TrackerClientError):
    """A ServiceNow tracker client error."""


class ServiceNowAsyncTrackerClient:
    """A ServiceNow async tracker client."""

    _configuration: ServiceNowConfiguration
    _servicenow_client: Client
    _message_formatter: ServiceNowReportMessageFormatter

    _attachments_list_description_item_markdown_template = Template('- [${name}](${url})')

    _incident_attachment_prefix: str = 'incident_'
    _comment_attachment_prefix: str = 'comment_'

    def __init__(
        self,
        configuration: ServiceNowConfiguration,
    ):
        """
        Initialize self.

        Args:
            configuration: a ServiceNow configuration
        """
        self._configuration = configuration
        self._servicenow_client = Client(
            address=cast(str, self._configuration.host),
            basic_auth=(
                cast(str, self._configuration.login),
                cast(str, self._configuration.password),
            ),
            use_ssl=cast(bool, self._configuration.use_ssl),
            verify_ssl=cast(bool, self._configuration.verify),
        )
        self._message_formatter = ServiceNowReportMessageFormatter()

    def _build_tracker_issue(
        self,
        issue_id: str,
        issue_url: str,
        closed: bool,
    ) -> TrackerIssue:
        return TrackerIssue(
            tracker_url=f'https://{cast(str, self._configuration.host)}',
            project=cast(str, self._configuration.host),
            issue_id=issue_id,
            issue_url=issue_url,
            closed=closed,
        )

    def _build_incident_url(
        self,
        sys_id: str,
    ) -> str:
        host = cast(str, self._configuration.host)
        return f'https://{host}/nav_to.do?uri=%2Fincident.do%3Fsys_id%3D{sys_id}'  # noqa: WPS323

    async def get_tracker_issue(
        self,
        issue_id: str,
    ) -> Optional[TrackerIssue]:
        """
        Get a tracker issue.

        Args:
            issue_id: an issue id

        Returns:
            The issue if it exists, else None
        """
        incident_data = await self._get_servicenow_incident(
            incident_id=issue_id,
        )
        if incident_data is None:
            return None
        issue_url = self._build_incident_url(
            sys_id=incident_data['sys_id'],
        )
        return self._build_tracker_issue(
            issue_id=issue_id,
            issue_url=issue_url,
            closed=incident_data['state'].value.lower() == 'closed',
        )

    async def _get_servicenow_incident(
        self,
        incident_id: str,
    ) -> Optional[Record]:
        async with IncidentModel(
            self._servicenow_client,
            table_name='incident',
        ) as incident_model:
            try:
                incident_response = await incident_model.get_one(IncidentModel.sys_id == incident_id)
            except AiosnowException:
                return None
        return _as_dict(incident_response.data)

    async def get_tracker_issue_comments(
        self,
        issue_id: str,
        exclude_comments: Optional[List[str]] = None,
    ) -> TrackerIssueComments:
        """
        Get a list of comments on an issue.

        Args:
            issue_id: an issue id
            exclude_comments: an optional list of comment to exclude

        Returns:
            The list of comments
        """
        incident_data = await self._get_servicenow_incident(
            incident_id=issue_id,
        )
        if incident_data is None:
            return []
        incident_sys_id = cast(str, incident_data['sys_id'])
        wrapped_items: List[RecordWrapper] = []
        wrapped_items += map(
            CommentRecordWrapper, await self._get_incident_journal_comments(
                incident_sys_id=incident_sys_id,
                exclude_comments=exclude_comments,
            ),
        )
        wrapped_items += map(
            AttachmentRecordWrapper, await self._get_incident_attachments(
                incident_sys_id=incident_sys_id,
                exclude_comments=exclude_comments,
            ),
        )
        wrapped_items = sorted(wrapped_items, key=_sort_key_wrapped_sys_created_on)
        return await self._extract_comments_from_records(
            records=wrapped_items,
        )

    async def _extract_comments_from_records(
        self,
        records: List[RecordWrapper],
    ) -> List[TrackerIssueComment]:
        comments = []
        for wrapped_item in records:
            if isinstance(wrapped_item, CommentRecordWrapper):
                comments.append(
                    self._extract_comment(
                        comment_data=wrapped_item.data,
                    ),
                )
            elif isinstance(wrapped_item, AttachmentRecordWrapper):
                comment = await self._extract_comment_attachment_from_record(
                    record=wrapped_item,
                )
                if comment.attachments:
                    comments.append(
                        comment,
                    )
        return comments

    async def _extract_comment_attachment_from_record(
        self,
        record: AttachmentRecordWrapper,
    ) -> TrackerIssueComment:
        comment = await self._extract_attachment(
            attachment_data=record.data,
        )
        attachments = {}
        for attachment_file_name, attachment in comment.attachments.items():
            is_incident_attachment = attachment_file_name.startswith(self._incident_attachment_prefix)
            is_comment_attachment = attachment_file_name.startswith(self._comment_attachment_prefix)
            if not is_incident_attachment and not is_comment_attachment:
                attachments[attachment_file_name] = attachment
        comment.attachments = attachments
        return comment

    def _exclude_from_list(
        self,
        records: List[Record],
        record_property: str,
        exclude: Set[str],
    ) -> List[Record]:
        return list(
            filter(
                lambda record: record[record_property] not in exclude,
                records,
            ),
        )

    async def _get_incident_journal_comments(
        self,
        incident_sys_id: str,
        exclude_comments: Optional[List[str]] = None,
    ) -> List[Record]:
        journal_condition = JournalModel.element_id.equals(incident_sys_id)
        journal_condition &= JournalModel.element.equals('comments')
        # limited pre-filter on query because API url might be too long
        for exclude_comment in (exclude_comments or [])[:_MODEL_NOT_EQUALS_LIMIT]:
            journal_condition &= JournalModel.sys_id.not_equals(exclude_comment)
        journal_query = select(journal_condition).order_asc(JournalModel.sys_created_on)
        async with JournalModel(
            self._servicenow_client,
            table_name='sys_journal_field',
        ) as journal_model:
            try:
                records = list(cast(Iterable[Record], await journal_model.get(journal_query)))
            except AiosnowException as journal_exception:
                raise ServiceNowTrackerClientError(
                    f'Unable to get journal comments for incident {incident_sys_id}',
                ) from journal_exception
        return self._exclude_from_list(
            records=records,
            record_property='sys_id',
            exclude=set(exclude_comments or []),
        )

    async def _get_incident_attachments(
        self,
        incident_sys_id: str,
        exclude_comments: Optional[List[str]] = None,
    ) -> List[Record]:
        attachment_condition = InMemoryAttachmentModel.table_name.equals('incident')
        attachment_condition &= InMemoryAttachmentModel.table_sys_id.equals(incident_sys_id)
        # limited pre-filter on query because API url might be too long
        for exclude_comment in (exclude_comments or [])[:_MODEL_NOT_EQUALS_LIMIT]:
            attachment_condition &= JournalModel.sys_id.not_equals(exclude_comment)
        attachment_query = select(attachment_condition).order_asc(InMemoryAttachmentModel.sys_created_on)
        async with InMemoryAttachmentModel(
            self._servicenow_client,
            table_name='incident',
        ) as attachment_model:
            try:
                records = list(cast(Iterable[Record], await attachment_model.get(attachment_query)))
            except AiosnowException as attachment_exception:
                raise ServiceNowTrackerClientError(
                    f'Unable to get attachments for incident {incident_sys_id}',
                ) from attachment_exception
        return self._exclude_from_list(
            records=records,
            record_property='sys_id',
            exclude=set(exclude_comments or []),
        )

    def _extract_comment(
        self,
        comment_data: Record,
    ) -> TrackerIssueComment:
        return TrackerIssueComment(
            created_at=comment_data['sys_created_on'],
            author=comment_data['sys_created_by'],
            comment_id=comment_data['sys_id'],
            body=comment_data['value'],
            attachments={},
        )

    async def _extract_attachment(
        self,
        attachment_data: Record,
    ) -> TrackerIssueComment:
        async with InMemoryAttachmentModel(
            self._servicenow_client,
            table_name='incident',
        ) as attachment_model:
            try:
                attachment_content = await attachment_model.download(
                    InMemoryAttachmentModel.sys_id == cast(str, attachment_data['sys_id']),
                )
            except AiosnowException as attachment_exception:
                raise ServiceNowTrackerClientError(
                    f'Unable to download attachment for incident {attachment_data["sys_id"]}',
                ) from attachment_exception
        file_name = cast(str, attachment_data['file_name'])
        return TrackerIssueComment(
            created_at=attachment_data['sys_created_on'],
            author=attachment_data['sys_created_by'],
            comment_id=attachment_data['sys_id'],
            body=f'Attachment:\n![{file_name}]({file_name})',
            attachments={
                file_name: TrackerAttachment(
                    filename=file_name,
                    mime_type=cast(str, attachment_data['content_type']),
                    content=attachment_content,
                ),
            },
        )

    async def send_report(
        self,
        report: Report,
    ) -> TrackerIssue:
        """
        Send a report to the tracker.

        Args:
            report: a report

        Returns:
            information about the sent report
        """
        short_description = self._message_formatter.format_report_title(
            report=report,
        )
        description = self._message_formatter.format_report_description(
            report=report,
        ) + self._get_attachments_list_description(
            title='Attachments:',
            item_template=self._attachments_list_description_item_markdown_template,
            file_name_prefix=self._incident_attachment_prefix,
            attachments=report.attachments,
        )
        markdown_description = ''
        description_attachment = None
        if len(description) > _TEXT_MAX_SIZE:
            description_attachment = self._build_external_description_attachment(
                name=f'report-{report.local_id.replace("#", "")}-description.md',
            )
            markdown_description = ReportMessageMarkdownFormatter().format_report_description(
                report=report,
            ) + self._get_attachments_list_description(
                title='**Attachments**:',
                item_template=self._attachments_list_description_item_markdown_template,
                file_name_prefix=self._incident_attachment_prefix,
                attachments=report.attachments,
            )
            report_copy = deepcopy(report)
            report_copy.description_html = (
                '<p>This report description is too big to fit into a ServiceNow issue. '
                + f'See attachment <a href="{description_attachment.url}">{description_attachment.original_name}</a> '
                + 'for more details.</p>'
            )
            description = self._replace_inline_attachments(
                attachments=[
                    description_attachment,
                ],
                file_name_prefix=self._incident_attachment_prefix,
                content=self._message_formatter.format_report_description(
                    report=report_copy,
                ) + self._get_attachments_list_description(
                    title='Attachments:',
                    item_template=self._attachments_list_description_item_markdown_template,
                    file_name_prefix=self._incident_attachment_prefix,
                    attachments=[
                        description_attachment,
                        *report.attachments,
                    ],
                ),
            )
        description, markdown_description = self._replace_attachments_references(
            attachments=report.attachments,
            file_name_prefix=self._incident_attachment_prefix,
            referencing_texts=[
                description,
                markdown_description,
            ],
        )
        attachments = report.attachments
        if description_attachment:
            description_attachment.data_loader = lambda: bytes(markdown_description, 'utf-8')
            description = self._replace_attachments_references(
                attachments=report.attachments,
                file_name_prefix=self._incident_attachment_prefix,
                referencing_texts=[
                    description,
                ],
            )[0]
            attachments = [
                description_attachment,
                *attachments,
            ]
        incident_data = await self._create_incident(
            short_description=short_description,
            description=description,
        )
        sys_id = incident_data['sys_id']
        await self._upload_attachments(
            attachments=attachments,
            file_name_prefix=self._incident_attachment_prefix,
            table_name='incident',
            record_sys_id=sys_id,
        )
        issue_url = self._build_incident_url(
            sys_id=sys_id,
        )
        return self._build_tracker_issue(
            issue_id=sys_id,
            issue_url=issue_url,
            closed=False,
        )

    def _build_external_description_attachment(
        self,
        name: str,
    ) -> Attachment:
        return Attachment(
            attachment_id=0,
            name=name,
            original_name=name,
            mime_type='text/markdown',
            size=0,
            url=f'http://tracker/external/{name}',
            data_loader=lambda: bytes('', 'utf-8'),
        )

    async def _create_incident(
        self,
        short_description: str,
        description: str,
    ) -> Record:
        async with IncidentModel(
            self._servicenow_client,
            table_name='incident',
        ) as incident_model:
            try:
                incident_response = await incident_model.create(
                    {
                        'description': description,
                        'short_description': short_description,
                    },
                )
            except AiosnowException as incident_create_exception:
                raise ServiceNowTrackerClientError('Unable to create incident') from incident_create_exception
        return _as_dict(incident_response.data)

    async def _upload_attachments(
        self,
        attachments: List[Attachment],
        file_name_prefix: str,
        table_name: str,
        record_sys_id: str,
    ) -> List[Tuple[str, str]]:
        uploaded_attachments = []
        host = cast(str, self._configuration.host)
        try:
            async with InMemoryAttachmentModel(
                self._servicenow_client,
                table_name=table_name,
            ) as attachment_model:
                for attachment in attachments:
                    attachment_response = await attachment_model.upload(
                        table_name=table_name,
                        record_sys_id=record_sys_id,
                        file_name=f'{file_name_prefix}{attachment.original_name}',
                        content_type=attachment.mime_type,
                        content=attachment.data,
                    )
                    sys_id = _as_dict(attachment_response.data)['sys_id']
                    view_url = f'https://{host}/sys_attachment.do?view=false&sys_id={sys_id}'
                    uploaded_attachments.append(
                        (
                            attachment.url,
                            view_url,
                        ),
                    )
        except AiosnowException as attachment_upload_exception:
            raise ServiceNowTrackerClientError(
                f'Unable to upload attachments to {table_name}',
            ) from attachment_upload_exception
        return uploaded_attachments

    def _get_attachments_list_description(
        self,
        title: str,
        item_template: Template,
        file_name_prefix: str,
        attachments: List[Attachment],
    ) -> str:
        attachments_lines = []
        if attachments:
            attachments_lines = [
                '',
                title,
            ]
            for attachment in attachments:
                attachments_lines.append(
                    item_template.substitute(
                        name=attachment.original_name,
                        url=attachment.url,
                    ),
                )
            attachments_lines.append('')
        return self._replace_inline_attachments(
            attachments=attachments,
            content='\n'.join(attachments_lines),
            file_name_prefix=file_name_prefix,
        )

    async def send_logs(
        self,
        tracker_issue: TrackerIssue,
        logs: List[Log],
    ) -> SendLogsResult:
        """
        Send logs to the tracker.

        Args:
            tracker_issue: information about the tracker issue
            logs: a list of comments

        Raises:
            ServiceNowTrackerClientError: if an error occurs

        Returns:
            information about the sent comments
        """
        data = _as_optional_dict(
            await self._get_servicenow_incident(
                incident_id=tracker_issue.issue_id,
            ),
        )
        if data is None:
            raise ServiceNowTrackerClientError(
                f'ServiceNow incident {tracker_issue.issue_id} not found',
            )
        incident_data = data
        tracker_comments = SendLogsResult(
            tracker_issue=tracker_issue,
            added_comments=[],
        )
        if incident_data['state'].value.lower() == 'closed':
            return tracker_comments
        for log in logs:
            comment_data, comment = await self._add_comment(
                incident_data=incident_data,
                log=log,
            )
            tracker_comments.added_comments.append(
                TrackerIssueComment(
                    created_at=comment_data['sys_created_on'],
                    author=comment_data['sys_created_by'],
                    comment_id=comment_data['sys_id'],
                    body=comment,
                    attachments={},
                ),
            )
        return tracker_comments

    async def _add_comment(
        self,
        incident_data: Record,
        log: Log,
    ) -> Tuple[Record, str]:
        incident_id = cast(str, incident_data['sys_id'])
        comment = self._message_formatter.format_log(
            log=log,
        ) + self._get_attachments_list_description(
            title='Attachments:',
            item_template=self._attachments_list_description_item_markdown_template,
            file_name_prefix=self._incident_attachment_prefix,
            attachments=log.attachments,
        )
        markdown_comment = ''
        comment_attachment = None
        if len(comment) > _TEXT_MAX_SIZE:
            comment_attachment = self._build_external_description_attachment(
                name=f'comment-{log.log_id}-description.md',
            )
            markdown_comment = ReportMessageMarkdownFormatter().format_log(
                log=log,
            ) + self._get_attachments_list_description(
                title='**Attachments**:',
                item_template=self._attachments_list_description_item_markdown_template,
                file_name_prefix=self._comment_attachment_prefix,
                attachments=log.attachments,
            )
            log_copy = deepcopy(log)
            log_copy.message_html = (
                '<p>This comment is too big to fit into a ServiceNow comment. '
                + f'See attachment <a href="{comment_attachment.url}">{comment_attachment.original_name}</a> '
                + 'for more details.</p>'
            )
            comment = self._replace_inline_attachments(
                attachments=[
                    comment_attachment,
                ],
                file_name_prefix=self._incident_attachment_prefix,
                content=self._message_formatter.format_log(
                    log=log_copy,
                ) + self._get_attachments_list_description(
                    title='Attachments:',
                    item_template=self._attachments_list_description_item_markdown_template,
                    file_name_prefix=self._comment_attachment_prefix,
                    attachments=[
                        comment_attachment,
                        *log.attachments,
                    ],
                ),
            )
        comment, markdown_comment = self._replace_attachments_references(
            attachments=log.attachments,
            file_name_prefix=self._comment_attachment_prefix,
            referencing_texts=[
                comment,
                markdown_comment,
            ],
        )
        attachments = log.attachments
        if comment_attachment:
            comment_attachment.data_loader = lambda: bytes(markdown_comment, 'utf-8')
            comment = self._replace_attachments_references(
                attachments=log.attachments,
                file_name_prefix=self._incident_attachment_prefix,
                referencing_texts=[
                    comment,
                ],
            )[0]
            attachments = [
                comment_attachment,
                *attachments,
            ]
        async with IncidentModel(
            self._servicenow_client,
            table_name='incident',
        ) as incident_model:
            try:
                await incident_model.update(
                    IncidentModel.sys_id == incident_id,
                    {
                        'comments': comment,
                    },
                )
            except AiosnowException as incident_update_exception:
                raise ServiceNowTrackerClientError(
                    f'Unable to add comment to incident number {incident_data["number"]}',
                ) from incident_update_exception
        comment_data = await self._get_incident_comment_id(
            incident_id=incident_id,
            comment=comment,
        )
        await self._upload_attachments(
            attachments=attachments,
            file_name_prefix=self._comment_attachment_prefix,
            table_name='incident',
            record_sys_id=incident_data['sys_id'],
        )
        return comment_data, comment

    def _replace_attachments_references(
        self,
        attachments: List[Attachment],
        file_name_prefix: str,
        referencing_texts: List[str],
    ) -> List[str]:
        return [
            self._replace_inline_attachments(
                attachments=attachments,
                content=text,
                file_name_prefix=file_name_prefix,
            )
            for text in referencing_texts
        ]

    def _replace_inline_attachments(
        self,
        attachments: List[Attachment],
        content: str,
        file_name_prefix: str,
    ) -> str:
        attachment_urls = {attachment.url for attachment in attachments}
        inline_attachments = _RE_INLINE_ATTACHMENT.findall(content)
        for match, attachment_name, url in inline_attachments:
            if url in attachment_urls:
                content = content.replace(match, f'{file_name_prefix}{attachment_name}')
        return content

    async def _get_incident_comment_id(
        self,
        incident_id: str,
        comment: str,
    ) -> Record:
        async with JournalModel(
            self._servicenow_client,
            table_name='sys_journal_field',
        ) as journal_model:
            condition = JournalModel.element_id.equals(incident_id)
            condition &= JournalModel.element.equals('comments')
            condition &= JournalModel.value.equals(
                self._escape_query_term_separator(
                    term=comment,
                ),
            )
            try:
                journal_response = await journal_model.get_one(condition)
            except AiosnowException as journal_comment_exception:
                raise ServiceNowTrackerClientError(
                    f'Unable to get comment from incident {incident_id}',
                ) from journal_comment_exception
            return _as_dict(journal_response.data)

    async def test(
        self,
    ) -> None:
        """
        Test the client.

        Raises:
            ServiceNowTrackerClientError: if the test failed
        """
        # The query itself does not really matter, we only want to be sure that we can perform a query.
        # If the credentials are not valid, the query will fail.
        async with UserModel(
            self._servicenow_client,
            table_name='sys_user',
        ) as user_model:
            try:
                await user_model.get_one(UserModel.user_name == self._configuration.login)
            except AiosnowException as user_exception:
                raise ServiceNowTrackerClientError('User not found') from user_exception

    def _escape_query_term_separator(
        self,
        term: str,
    ) -> str:
        return term.replace('^', '^^')


class ServiceNowTrackerClient(TrackerClient[ServiceNowConfiguration]):
    """A ServiceNow tracker client wrapping an async client."""

    _async_client: ServiceNowAsyncTrackerClient

    def __init__(
        self,
        configuration: ServiceNowConfiguration,
    ):
        """
        Initialize self.

        Args:
            configuration: a ServiceNow configuration
        """
        super().__init__(
            configuration=configuration,
        )
        self._async_client = ServiceNowAsyncTrackerClient(
            configuration=configuration,
        )

    @property
    def tracker_type(
        self,
    ) -> str:
        """
        Get the type of the tracker client.

        Returns:
            the type of the tracker client
        """
        return 'ServiceNow'

    def get_tracker_issue(
        self,
        issue_id: str,
    ) -> Optional[TrackerIssue]:
        """
        Get a tracker issue.

        Args:
            issue_id: an issue id

        Returns:
            The issue if it exists, else None
        """
        return asyncio.run(
            self._async_client.get_tracker_issue(
                issue_id=issue_id,
            ),
        )

    def get_tracker_issue_comments(
        self,
        issue_id: str,
        exclude_comments: Optional[List[str]] = None,
    ) -> TrackerIssueComments:
        """
        Get a list of comments on an issue.

        Args:
            issue_id: an issue id
            exclude_comments: an optional list of comment to exclude

        Returns:
            The list of comments
        """
        return asyncio.run(
            self._async_client.get_tracker_issue_comments(
                issue_id=issue_id,
                exclude_comments=exclude_comments,
            ),
        )

    def send_report(
        self,
        report: Report,
    ) -> TrackerIssue:
        """
        Send a report to the tracker.

        Args:
            report: a report

        Returns:
            information about the sent report
        """
        return asyncio.run(
            self._async_client.send_report(
                report=report,
            ),
        )

    def send_logs(
        self,
        tracker_issue: TrackerIssue,
        logs: List[Log],
    ) -> SendLogsResult:
        """
        Send logs to the tracker.

        Args:
            tracker_issue: information about the tracker issue
            logs: a list of comments

        Returns:
            information about the sent comments
        """
        return asyncio.run(
            self._async_client.send_logs(
                tracker_issue=tracker_issue,
                logs=logs,
            ),
        )

    def test(
        self,
    ) -> None:
        """Test the client."""
        asyncio.run(self._async_client.test())


def _sort_key_wrapped_sys_created_on(
    wrapped: RecordWrapper,
) -> datetime.datetime:
    return cast(datetime.datetime, wrapped.data['sys_created_on'])


ResponseDataType = Union[List[Any], Record, bytes]


def _as_optional_dict(data: Optional[ResponseDataType]) -> Optional[Record]:
    if data is not None and not isinstance(data, dict):
        raise ServiceNowTrackerClientError(f'Expecting dict, got {type(data)}')
    return data


def _as_dict(data: Optional[ResponseDataType]) -> Record:
    if not isinstance(data, dict):
        raise ServiceNowTrackerClientError(f'Expecting non-empty dict, got {type(data)}')
    return data



"""Models and functions used for JSON schema."""
import json
from functools import partial
from string import Template
from typing import Any, Callable, Dict, List

from ywh2bt.core.schema.error import SchemaError
from ywh2bt.core.schema.json import root_configuration_as_json_schema

Json = Dict[str, Any]


def root_configuration_as_markdown() -> str:
    """
    Get a markdown schema for RootConfiguration.

    Raises:
        SchemaError: if an error occurred during the creation of the schema

    Returns:
        the schema
    """
    json_schema_str = root_configuration_as_json_schema()
    try:
        schema: Json = json.loads(json_schema_str)
    except TypeError as load_error:
        raise SchemaError('JSON load error') from load_error
    return '\n'.join(
        _schema_to_markdown(
            schema=schema,
        ),
    )


def _schema_to_markdown(
    schema: Json,
) -> List[str]:
    ref = schema.get('$ref')
    if ref:
        return [
            Template('* **Reference**: [`${reference}`](#${reference})').substitute(
                reference=ref.replace('#/definitions/', ''),
            ),
        ]
    any_of = schema.get('anyOf')
    if any_of:
        return _extract_any_of(
            any_of=any_of,
        )
    output = _extract_attributes(
        schema=schema,
        attributes={
            'title': Template('* **Title**: ${value}'),
            'description': Template('* **Description**: ${value}'),
            'type': Template('* **Type**: `${value}`'),
            'const': Template('* **Required constant value**: `${value}`'),
            'default': Template('* **Default value**: `${value}`'),
        },
    )
    output.extend(
        _extract_properties(
            schema=schema,
        ),
    )
    output.extend(
        _extract_properties_patterns(
            schema=schema,
        ),
    )
    output.extend(
        _extract_definitions(
            schema=schema,
        ),
    )
    output.extend(
        _extract_items(
            schema=schema,
        ),
    )

    return output


def _extract_attributes(
    schema: Json,
    attributes: Dict[str, Template],
) -> List[str]:
    output = []
    for attribute, template in attributes.items():
        value = schema.get(attribute)
        if value is not None:
            output.append(
                template.substitute(
                    value=value,
                ),
            )
    return output


def _extract_properties(
    schema: Json,
) -> List[str]:
    required = schema.get('required', [])
    properties = []
    for property_name, property_schema in schema.get('properties', {}).items():
        properties.append(f'- `{property_name}`')
        properties.extend(
            _prefix_lines(
                '    ',
                _schema_to_markdown(
                    schema=property_schema,
                ),
            ),
        )
        if property_name in required:
            properties.append('    * **Required**: True')
    if not properties:
        return []
    return [
        '* **Properties**:',
        *_prefix_lines(
            '    ',
            properties,
        ),
    ]


def _extract_properties_patterns(
    schema: Json,
) -> List[str]:
    return _extract_list(
        schema=schema,
        attribute_name='patternProperties',
        block_label='* **Properties patterns**:',
        item_name_template=Template('- **Pattern**: `${item_name}`'),
    )


def _extract_definitions(
    schema: Json,
) -> List[str]:
    return _extract_list(
        schema=schema,
        attribute_name='definitions',
        block_label='* **Definitions**:',
        item_name_template=Template('- <a name="${item_name}">`${item_name}`</a>:'),
    )


def _extract_items(
    schema: Json,
) -> List[str]:
    items = schema.get('items')
    if not items:
        return []
    return [
        '* **Items**:',
        *_prefix_lines(
            '    ',
            _schema_to_markdown(
                schema=items,
            ),
        ),
    ]


def _extract_list(
    schema: Json,
    attribute_name: str,
    block_label: str,
    item_name_template: Template,
) -> List[str]:
    items = []
    for item_name, item_schema in schema.get(attribute_name, {}).items():
        items.append(item_name_template.substitute(item_name=item_name))
        items.extend(
            _prefix_lines(
                '    ',
                _schema_to_markdown(
                    schema=item_schema,
                ),
            ),
        )
    if not items:
        return []
    return [
        block_label,
        *_prefix_lines(
            '    ',
            items,
        ),
    ]


def _extract_any_of(
    any_of: List[Json],
) -> List[str]:
    output = [
        '* Any of:',
    ]
    for any_of_entry in any_of:
        output.extend(
            _prefix_lines(
                '    ',
                list(_schema_to_markdown(
                    schema=any_of_entry,
                )),
            ),
        )
    return output


def _prefix(
    prefix: str,
    string: str,
) -> str:
    return f'{prefix}{string}'


def _prefix_lines(
    prefix: str,
    lines: List[str],
) -> List[str]:
    return list(map(
        _prefixer(prefix),
        lines,
    ))


def _prefixer(
    prefix: str,
) -> Callable[[str], str]:
    return partial(
        _prefix,
        prefix,
    )



"""Models and functions used for root configurations GUI."""
from __future__ import annotations

from typing import Any, List, Optional, Tuple, cast

from PySide2.QtCore import QFileInfo, QSettings, Qt, Signal
from PySide2.QtWidgets import QBoxLayout, QFileDialog, QInputDialog, QMessageBox, QTabWidget, QWidget

from ywh2bt.core.core import AVAILABLE_FORMATS
from ywh2bt.core.exceptions import CoreException
from ywh2bt.gui.dialog.error import ErrorDialogMixin
from ywh2bt.gui.dialog.file import FileFormatDialogFilters
from ywh2bt.gui.widgets.root_configuration_entry import RootConfigurationEntry
from ywh2bt.gui.widgets.root_configuration_widget import RootConfigurationWidget
from ywh2bt.gui.widgets.typing import as_signal_instance


class RootConfigurationsWidget(QWidget, ErrorDialogMixin):  # noqa: WPS214
    """
    Root configurations GUI.

    Display and edit multiple configurations in visual tabs.
    """

    entry_modified: Signal = Signal(RootConfigurationEntry)
    entry_saved: Signal = Signal(RootConfigurationEntry)
    current_entry_changed: Signal = Signal(RootConfigurationEntry)

    _file_format_dialog_filters: FileFormatDialogFilters
    _settings: QSettings
    _empty_widget: QWidget
    _tab_widget: QTabWidget

    _entries: List[RootConfigurationEntry]

    def __init__(
        self,
        settings: QSettings,
        empty_widget: QWidget,
        parent: Optional[QWidget] = None,
        *args: Any,
        **kwargs: Any,
    ):
        """
        Initialize self.

        Args:
            settings: application settings
            empty_widget: a widget to display if no configurations are currently being edited
            parent: a parent widget
            args: extra arguments
            kwargs: extra keyword arguments
        """
        super().__init__(
            parent,
            *args,
            **kwargs,
        )
        self._settings = settings
        self._entries = []
        self._file_format_dialog_filters = FileFormatDialogFilters(
            formats=AVAILABLE_FORMATS,
        )
        self._init_ui(
            empty_widget=empty_widget,
        )

    def _init_ui(
        self,
        empty_widget: QWidget,
    ) -> None:
        layout = QBoxLayout(
            QBoxLayout.LeftToRight,
            self,
        )

        self._empty_widget = empty_widget
        self._tab_widget = self._create_tab_widget()

        self._auto_show_hide_tabs()

        layout.addWidget(self._empty_widget, Qt.AlignCenter)
        layout.addWidget(self._tab_widget)
        layout.setMargin(0)

    def _create_tab_widget(
        self,
    ) -> QTabWidget:
        widget = QTabWidget(self)
        widget.setTabsClosable(True)
        as_signal_instance(widget.currentChanged).connect(
            self._on_current_tab_changed,
        )
        as_signal_instance(widget.tabCloseRequested).connect(
            self._on_tab_close_requested,
        )
        return widget

    def _on_current_tab_changed(
        self,
        tab_index: int,
    ) -> None:
        entry = None
        if tab_index >= 0:
            entry = self._entries[tab_index]
        as_signal_instance(self.current_entry_changed).emit(
            entry,
        )

    def _on_tab_close_requested(
        self,
        tab_index: int,
    ) -> None:
        entry = self._entries[tab_index]
        if not entry.has_changed():
            self._on_entry_close_accepted(
                entry=entry,
            )
            return
        reply = QMessageBox.question(
            self,
            f'Close {entry.name}',
            'This configuration has unsaved changes. Close anyway?',
            QMessageBox.Yes | QMessageBox.No,
            QMessageBox.No,
        )
        if reply == QMessageBox.Yes:
            self._on_entry_close_accepted(
                entry=entry,
            )

    def _on_entry_close_accepted(
        self,
        entry: RootConfigurationEntry,
    ) -> None:
        tab_index = self._get_tab_index_for_entry(
            entry=entry,
        )
        self._tab_widget.removeTab(tab_index)
        self._entries.pop(tab_index)
        self._auto_show_hide_tabs()

    def _auto_show_hide_tabs(
        self,
    ) -> None:
        if self._tab_widget.count() == 0:
            self._tab_widget.hide()
            self._empty_widget.show()
        else:
            self._tab_widget.show()
            self._empty_widget.hide()

    def create_new_configuration(
        self,
    ) -> None:
        """Create a new configuration."""
        selected_format, ok = QInputDialog().getItem(
            self,
            'Select format',
            'Format:',
            list(AVAILABLE_FORMATS.keys()),
            0,
            False,  # noqa: WPS425
        )
        if not ok:
            return
        self.add_entry(
            entry=RootConfigurationEntry(
                name='New configuration',
                raw='',
                original_raw='',
                raw_format=selected_format,
            ),
        )

    def prompt_open_configuration(
        self,
    ) -> None:
        """Prompt for opening an existing configuration."""
        open_file = self._prompt_open_file()
        if not open_file:
            return
        file_path, file_format = open_file
        self._try_open_configuration(
            file_info=QFileInfo(file_path),
            file_format=file_format,
        )

    def reload_current_configuration(
        self,
    ) -> None:
        """Reload the current configuration."""
        entry = self.get_current_entry()
        if not entry.file:
            return
        reply = QMessageBox.question(
            self,
            f'Reload {entry.name}',
            f'Reload {entry.file.info.filePath()} from disk?\nUnsaved changes will be lost.',
            QMessageBox.Yes | QMessageBox.No,
            QMessageBox.No,
        )
        if reply == QMessageBox.No:
            return
        self._try_open_configuration(
            file_info=entry.file.info,
            file_format=entry.file.file_format,
        )

    def _prompt_open_file(
        self,
    ) -> Optional[Tuple[str, str]]:
        last_opened_dir = self._get_last_opened_dir()
        file_path, chosen_filter = QFileDialog.getOpenFileName(
            self,
            'Open file',
            last_opened_dir,
            self._file_format_dialog_filters.get_filters_string(),
        )
        if not file_path:
            return None
        chosen_format = self._file_format_dialog_filters.get_format_name(
            filter_string=chosen_filter,
        )
        if chosen_format is None:
            return None
        self._settings.setValue('last_opened_file', file_path)
        self._settings.setValue('last_opened_file_format', chosen_format)
        return file_path, chosen_format

    def _try_open_configuration(
        self,
        file_info: QFileInfo,
        file_format: str,
    ) -> None:
        try:
            entry = RootConfigurationEntry.from_file(
                file_info=file_info,
                file_format=file_format,
            )
        except CoreException as e:
            self.show_exception_dialog(
                parent=self,
                text='Open error',
                informative_text='Unable to open configuration',
                exception=e,
            )
            return
        self.add_entry(
            entry=entry,
        )

    def _get_last_opened_dir(
        self,
    ) -> str:
        return QFileInfo(
            cast(str, self._settings.value('last_opened_file')),
        ).dir().path()

    def add_entry(
        self,
        entry: RootConfigurationEntry,
    ) -> None:
        """
        Add a configuration entry to be edited.

        Args:
            entry: a configuration entry
        """
        existing_entry = self._get_same_file_entry(
            entry=entry,
        )
        if existing_entry:
            existing_entry.configuration = entry.configuration
            tab_index = self.refresh_entry(
                entry=existing_entry,
            )
        else:
            widget = RootConfigurationWidget(
                settings=self._settings,
                parent=self,
                entry=entry,
            )
            as_signal_instance(widget.root_configuration_modified).connect(
                self._on_entry_modified,
            )
            as_signal_instance(widget.root_configuration_saved).connect(
                self._on_entry_saved,
            )
            as_signal_instance(widget.root_configuration_saved_as).connect(
                self._on_entry_saved_as,
            )
            self._entries.append(entry)
            tab_index = self._tab_widget.addTab(
                widget,
                _tab_title(
                    entry=entry,
                ),
            )
            self._update_tab_text(
                entry=entry,
            )
        self._tab_widget.setCurrentIndex(tab_index)
        self._auto_show_hide_tabs()

    def has_changed_entries(
        self,
    ) -> bool:
        """
        Check if any opened entry has changed.

        Returns:
            True if any entry has changed.
        """
        return any(entry.has_changed() for entry in self._entries)

    def refresh_entry(
        self,
        entry: RootConfigurationEntry,
    ) -> int:
        """
        Refresh the display of a configuration entry.

        Args:
            entry: a configuration entry

        Returns:
            The index at which this configuration is placed in the edited list
        """
        tab_index = self._get_tab_index_for_entry(
            entry=entry,
        )
        widget = cast(RootConfigurationWidget, self._tab_widget.widget(tab_index))
        widget.set_entry(
            entry=entry,
        )
        self._update_tab_text(
            entry=entry,
        )
        as_signal_instance(self.entry_modified).emit(
            entry,
        )
        return tab_index

    def save_current_entry(
        self,
    ) -> None:
        """Save the current entry configuration."""
        self._get_current_entry_widget().save()

    def save_as_current_entry(
        self,
    ) -> None:
        """Save the current entry configuration in a new file."""
        self._get_current_entry_widget().save_as()

    def test_current_entry(
        self,
    ) -> None:
        """Synchronize the current entry configuration."""
        self._get_current_entry_widget().test()

    def synchronize_current_entry(
        self,
    ) -> None:
        """Synchronize the current entry configuration."""
        self._get_current_entry_widget().synchronize()

    def _get_current_entry_widget(
        self,
    ) -> RootConfigurationWidget:
        return cast(RootConfigurationWidget, self._tab_widget.widget(
            self._tab_widget.currentIndex(),
        ))

    def _get_same_file_entry(
        self,
        entry: RootConfigurationEntry,
    ) -> Optional[RootConfigurationEntry]:
        if not entry.file:
            return None
        for existing_entry in self._entries:
            if existing_entry.file and existing_entry.file.info == entry.file.info:
                return existing_entry
        return None

    def _on_entry_modified(
        self,
        entry: RootConfigurationEntry,
    ) -> None:
        self._update_tab_text(
            entry=entry,
        )
        as_signal_instance(self.entry_modified).emit(
            entry,
        )

    def _on_entry_saved(
        self,
        entry: RootConfigurationEntry,
    ) -> None:
        as_signal_instance(self.entry_saved).emit(
            entry,
        )

    def _on_entry_saved_as(
        self,
        file_path: str,
        file_format: str,
    ) -> None:
        self._try_open_configuration(
            file_info=QFileInfo(file_path),
            file_format=file_format,
        )

    def _update_tab_text(
        self,
        entry: RootConfigurationEntry,
    ) -> None:
        color = Qt.black
        if not entry.configuration or not entry.configuration.is_valid():
            color = Qt.red
        tab_index = self._get_tab_index_for_entry(
            entry=entry,
        )
        self._tab_widget.tabBar().setTabTextColor(
            tab_index,
            color,
        )
        self._tab_widget.setTabText(
            tab_index,
            _tab_title(
                entry=entry,
            ),
        )

    def get_current_entry(
        self,
    ) -> RootConfigurationEntry:
        """
        Get the currently displayed configuration entry.

        Returns:
            The current configuration entry
        """
        tab_index = self._tab_widget.currentIndex()
        return self._entries[tab_index]

    def _get_tab_index_for_entry(
        self,
        entry: RootConfigurationEntry,
    ) -> int:
        return self._entries.index(entry)


def _tab_title(
    entry: RootConfigurationEntry,
) -> str:
    tab_title = entry.name
    if entry.has_changed():
        tab_title = f'* {tab_title}'
    return tab_title



"""Models and functions related to tabs (tab-widget, ...)."""
from typing import Any, Optional

from ywh2bt.core.configuration.yeswehack import Program


def object_to_tab_title(
    obj: Any,
) -> Optional[str]:
    """
    Extract from the given object a string to be used a tab title.

    Args:
        obj: an object

    Returns:
        The title if it could have been extracted; otherwise None
    """
    if isinstance(obj, Program):
        return obj.slug
    return None



"""Models and functions used in class subtyping and automatic casting."""
from __future__ import annotations

from abc import ABCMeta
from typing import Any, Dict, Optional, Tuple, Type, cast

from ywh2bt.core.configuration.error import BaseAttributeError

SUBTYPE_MANAGER_ATTR_NAME = '__subtype_manager'


class SubtypeError(BaseAttributeError):
    """An error related to a subtype."""


class Subtypable:
    """Base class for all subtypes."""


class SubtypableMetaclass(ABCMeta):
    """
    A metaclass for subtypes.

    Used to determine a subtype class from a string type.

    Examples:
        class CustomType(Subtypable):
            pass

        SubtypableMetaclass.register_subtype('custom-type', CustomType)
        instance = Subtypable(type='custom-type')
        assert isinstance(instance, CustomType)  # => True

    """

    def _get_subtype_manager(
        cls,
    ) -> SubtypableManager:
        return cast(
            SubtypableManager,
            getattr(cls, SUBTYPE_MANAGER_ATTR_NAME),
        )

    def get_registered_subtypes(
        cls,
    ) -> Dict[str, Type[Subtypable]]:
        """
        Get the registered subtypes.

        Returns:
            the registered subtypes
        """
        return cls._get_subtype_manager().get_registered_subtypes()

    def register_subtype(
        cls,
        subtype_name: str,
        subtype_class: Type[Subtypable],
    ) -> None:
        """
        Associate a type name with a subtype class.

        Args:
            subtype_name: the type name
            subtype_class: the subtype class

        # noqa: DAR101 mcs
        """
        cls._get_subtype_manager().register_subtype(
            subtype_name=subtype_name,
            subtype_class=subtype_class,
        )

    def get_subtype_name(
        cls,
        subtype_class: Type[Any],
    ) -> Optional[str]:
        """
        Get a subtype class from a type name.

        Args:
            subtype_class: a type name

        Returns:
            A subtype class

        # noqa: DAR101 mcs
        """
        return cls._get_subtype_manager().get_subtype_name(
            subtype_class=subtype_class,
        )

    def __new__(
        mcs,
        name: str,
        bases: Tuple[Type[Any], ...],
        namespace: Dict[str, Any],
    ) -> SubtypableMetaclass:
        """
        Call to create a new subtype class of class cls.

        Args:
            name: a name for the new class
            bases: class bases for the new class
            namespace: properties and attributes of the new class

        Returns:
            The new class
        """
        if Subtypable in bases:
            subtype_manager = SubtypableManager()
            namespace[SUBTYPE_MANAGER_ATTR_NAME] = subtype_manager

        return super(SubtypableMetaclass, mcs).__new__(mcs, name, bases, namespace)  # noqa: WPS608

    def __call__(
        cls,
        *args: Any,
        **kwargs: Any,
    ) -> Subtypable:
        """
        Call the subtypable type as a function.

        Shorthand for `Subtypable.__call__(*args, **kwargs)`.

        Args:
            args: an arguments list
            kwargs: keyword arguments

        Raises:
            SubtypeError: if the subtype is not found in the registered subtypes

        Returns:
            A subtype instance

        # noqa: DAR101 cls
        """
        new_cls = cls
        subtypes = getattr(cls, 'get_registered_subtypes')()  # noqa: B009
        if Subtypable in cls.__bases__:
            subtype_name = kwargs.pop('type', None)
            if subtype_name:
                if subtype_name in subtypes:
                    new_cls = subtypes[subtype_name]  # noqa: WPS529
                else:
                    raise SubtypeError(
                        message=f'Subtype {repr(subtype_name)} is not supported.',
                        context=None,
                    )
            else:
                raise SubtypeError(
                    message='No subtype provided.',
                    context=None,
                )
        obj = cls.__new__(new_cls, *args, **kwargs)  # type: ignore
        obj.__init__(*args, **kwargs)  # type: ignore  # noqa: WPS609
        return cast(Subtypable, obj)


class SubtypableManager:
    """A class for handling the subclasses of a subtypable class."""

    _subtypes: Dict[str, Type[Subtypable]]

    def __init__(
        self,
    ) -> None:
        """Initialize self."""
        self._subtypes = {}

    def get_registered_subtypes(
        self,
    ) -> Dict[str, Type[Subtypable]]:
        """
        Get the registered subtypes.

        Returns:
            the registered subtypes
        """
        return self._subtypes.copy()

    def register_subtype(
        self,
        subtype_name: str,
        subtype_class: Type[Subtypable],
    ) -> None:
        """
        Associate a type name with a subtype class.

        Args:
            subtype_name: the type name
            subtype_class: the subtype class

        Raises:
            SubtypeError: if the class is not a subclass of Subtypable

        # noqa: DAR101 mcs
        """
        if not issubclass(subtype_class, Subtypable):
            raise SubtypeError(
                message=f'Class {subtype_class} is not a subclass of {Subtypable}.',
                context=None,
            )
        self._subtypes[subtype_name] = subtype_class

    def get_subtype_name(
        self,
        subtype_class: Type[Any],
    ) -> Optional[str]:
        """
        Get a subtype class from a type name.

        Args:
            subtype_class: a type name

        Returns:
            A subtype class

        # noqa: DAR101 mcs
        """
        for registered_name, registered_class in self._subtypes.items():
            if registered_class is subtype_class:
                return registered_name
        return None



"""Models and functions used for observing synchronisation between YesWeHack and trackers."""
from abc import ABC
from dataclasses import dataclass
from typing import (
    List,
    Optional,
    Tuple,
)

from ywh2bt.core.api.models.report import Report
from ywh2bt.core.api.tracker import TrackerIssue
from ywh2bt.core.configuration.root import RootConfiguration
from ywh2bt.core.configuration.yeswehack import YesWeHackConfiguration
from ywh2bt.core.listener import (
    Event,
    Listener,
)


@dataclass
class SynchronizerEvent(Event, ABC):
    """A base event sent by a Synchronizer."""


@dataclass
class SynchronizerStartEvent(SynchronizerEvent):
    """An event sent when a synchronization starts."""

    configuration: RootConfiguration


@dataclass
class SynchronizerEndEvent(SynchronizerEvent):
    """An event sent when a synchronization ends."""

    configuration: RootConfiguration


@dataclass
class SynchronizerStartFetchReportsEvent(SynchronizerEvent):
    """An event sent when fetching of reports starts."""

    configuration: RootConfiguration
    yeswehack_name: str
    yeswehack_configuration: YesWeHackConfiguration
    program_slug: str


@dataclass
class SynchronizerEndFetchReportsEvent(SynchronizerEvent):
    """An event sent when fetching of reports ends."""

    configuration: RootConfiguration
    yeswehack_name: str
    yeswehack_configuration: YesWeHackConfiguration
    program_slug: str
    reports: List[Report]


@dataclass
class SynchronizerStartSendReportEvent(SynchronizerEvent):
    """An event sent when a synchronization of a report starts."""

    configuration: RootConfiguration
    yeswehack_name: str
    yeswehack_configuration: YesWeHackConfiguration
    program_slug: str
    tracker_name: str
    report: Report


@dataclass
class SynchronizerEndSendReportEvent(SynchronizerEvent):
    """An event sent when a synchronization of a report ends."""

    configuration: RootConfiguration
    yeswehack_name: str
    yeswehack_configuration: YesWeHackConfiguration
    program_slug: str
    tracker_name: str
    report: Report
    is_created_issue: bool
    is_existing_issue: bool
    new_report_status: Optional[Tuple[str, str]]
    tracking_status_updated: bool
    tracker_issue: Optional[TrackerIssue]
    issue_added_comments: List[str]
    report_added_comments: List[str]


class SynchronizerListener(Listener[SynchronizerEvent], ABC):
    """A listener receiving events from a Synchronizer."""


class NoOpSynchronizerListener(SynchronizerListener):
    """A listener receiving events from a Synchronizer and doing nothing."""

    def on_event(
        self,
        event: SynchronizerEvent,
    ) -> None:
        """
        Receive an event and do nothing.

        Args:
            event: an event
        """



"""Models used for the root configuration."""
from typing import Any, Dict, Optional, Union, cast

from ywh2bt.core.configuration.attribute import Attribute, AttributesContainer
from ywh2bt.core.configuration.error import AttributesError
from ywh2bt.core.configuration.tracker import TrackerConfiguration, Trackers
from ywh2bt.core.configuration.validator import not_empty_validator
from ywh2bt.core.configuration.yeswehack import Programs, YesWeHackConfigurations

BugtrackersAttributeType = Union[
    Dict[str, TrackerConfiguration],
    Optional[Trackers],
    Attribute[Trackers],
]
BugtrackersInitType = Union[
    Optional[Trackers],
    Optional[Dict[str, TrackerConfiguration]],
    Optional[Dict[str, Any]],
]
YesWeHackAttributeType = Union[
    Dict[str, Any],
    Optional[YesWeHackConfigurations],
    Attribute[YesWeHackConfigurations],
]
YesWeHackInitType = Union[
    Optional[YesWeHackConfigurations],
    Optional[Dict[str, Any]],
]


class RootConfiguration(AttributesContainer):
    """A root configuration."""

    bugtrackers: BugtrackersAttributeType = Attribute.create(
        value_type=Trackers,
        short_description='Trackers',
        description='Trackers to be synchronized with YesWeHack',
        required=True,
        validator=not_empty_validator,
    )
    yeswehack: YesWeHackAttributeType = Attribute.create(
        value_type=YesWeHackConfigurations,
        short_description='YesWeHack',
        required=True,
        validator=not_empty_validator,
    )

    def __init__(
        self,
        bugtrackers: BugtrackersInitType = None,
        yeswehack: YesWeHackInitType = None,
        **kwargs: Any,
    ):
        """
        Initialize the configuration.

        Args:
            bugtrackers: a configuration for the bugtrackers
            yeswehack: a configuration for YesWeHack
            kwargs: keyword arguments
        """
        super().__init__(**kwargs)
        self.bugtrackers = bugtrackers
        self.yeswehack = yeswehack

    def validate(self) -> None:
        """
        Validate the configuration.

        Raises:
            AttributesError: if the configuration is invalid

        # noqa: DAR401
        # noqa: DAR402
        """
        super_error: Optional[AttributesError] = None
        try:
            super().validate()
        except AttributesError as e:
            super_error = e

        errors = {}
        if self.yeswehack:
            errors.update(
                self._ensure_declared_trackers(
                    trackers=cast(Trackers, self.bugtrackers or {}),
                    yeswehack=cast(YesWeHackConfigurations, self.yeswehack or {}),
                ),
            )

        if errors:
            if not super_error:
                super_error = AttributesError(
                    message='Validation error',
                    errors={},
                    context=self,
                )
            super_error.errors.update(errors)

        if super_error:
            raise super_error

    def _ensure_declared_trackers(
        self,
        trackers: Trackers,
        yeswehack: YesWeHackConfigurations,
    ) -> Dict[str, AttributesError]:
        errors = {}
        for yeswehack_name, yeswehack_config in yeswehack.items():
            config_errors = self._ensure_bugtrackers_name(
                trackers=trackers,
                programs=cast(Programs, yeswehack_config.programs or {}),
            )
            for key, error in config_errors.items():
                errors[f'yeswehack.{yeswehack_name}.{key}'] = error
        return errors

    def _ensure_bugtrackers_name(
        self,
        trackers: Trackers,
        programs: Programs,
    ) -> Dict[str, AttributesError]:
        errors = {}
        for i, program in enumerate(programs):
            if not program.bugtrackers_name:
                continue
            for j, bugtracker_name in enumerate(program.bugtrackers_name):
                if bugtracker_name not in trackers:
                    errors[f'programs[{i}].bugtrackers_name[{j}]'] = AttributesError(
                        message=f'Bugtracker not {repr(bugtracker_name)} declared in bugtrackers',
                        errors={},
                        context=self,
                    )
        return errors



"""Models and functions used for YesWeHack platform data access."""
from json import JSONDecodeError
from typing import (
    Any,
    Dict,
    List,
    Optional,
    cast,
)
from urllib.parse import urlsplit

import requests
from yeswehack.api import (
    Attachment as YesWeHackRawApiAttachment,
    Log as YesWeHackRawApiLog,
    Report as YesWeHackRawApiReport,
    YesWeHack as YesWeHackRawApiClient,
)
from yeswehack.exceptions import APIError as YesWeHackRawAPiError

from ywh2bt.core.api.client import TestableApiClient
from ywh2bt.core.api.mapping import (
    MappingContext,
    map_raw_attachment,
    map_raw_logs,
    map_raw_report,
)
from ywh2bt.core.api.models.report import (
    Attachment,
    Log,
    Report,
)
from ywh2bt.core.configuration.yeswehack import (
    YesWeHackConfiguration,
)
from ywh2bt.core.exceptions import CoreException
from ywh2bt.version import __VERSION__


class YesWeHackApiClientError(CoreException):
    """A YesWeHack API client error."""


class YesWeHackApiClient(TestableApiClient):
    """A YesWeHack API client."""

    _configuration: YesWeHackConfiguration
    _raw_client: YesWeHackRawApiClient
    _yeswehack_domain: str
    _logged_in: bool

    def __init__(
        self,
        configuration: YesWeHackConfiguration,
    ):
        """
        Initialize self.

        Args:
            configuration: a YesWeHack configuration
        """
        self._configuration = configuration
        self._raw_client = self._build_raw_client()
        self._logged_in = False
        self._yeswehack_domain = self._extract_yeswehack_domain(
            url=cast(str, self._configuration.api_url),
        )

    def test(
        self,
    ) -> None:
        """Test the client."""
        self._ensure_login()

    def _build_raw_client(
        self,
    ) -> YesWeHackRawApiClient:
        configuration = self._configuration
        try:
            client = YesWeHackRawApiClient(
                api_url=self._normalize_api_url(
                    api_url=cast(str, configuration.api_url),
                ),
                pat=configuration.pat,
                verify=configuration.verify,
                lazy=True,
                headers={
                    'User-Agent': f'ywh2bt/{__VERSION__}',
                },
            )
        except (YesWeHackRawAPiError, requests.RequestException) as e:
            raise YesWeHackApiClientError('Unable to initialize YesWeHack API client') from e
        return client

    def _normalize_api_url(
        self,
        api_url: str,
    ) -> str:
        return api_url[:-1] if api_url[-1] == '/' else api_url

    def _extract_yeswehack_domain(
        self,
        url: str,
    ) -> str:
        result = urlsplit(url)
        # only keep 'xxx.yyy.zzz' from 'www.xxx.yyy.zzz'
        _, domain = result.netloc.split('.', 1)
        return domain

    def _ensure_login(
        self,
    ) -> None:
        if not self._logged_in:
            try:
                success = self._raw_client.login()
            except (YesWeHackRawAPiError, requests.RequestException) as e:
                raise YesWeHackApiClientError('Unable to log in with YesWeHack API client') from e
            self._logged_in = success

    def get_program_reports(
        self,
        slug: str,
        filters: Optional[Dict[str, Any]] = None,
    ) -> List[Report]:
        """
        Get reports for the program.

        Args:
            slug: a program slug
            filters: a report filter

        Returns:
            the list of reports

        Raises:
            YesWeHackApiClientError: if an error occurred when getting the reports
        """
        self._ensure_login()
        try:
            raw_reports = self._raw_client.get_reports(
                program=slug,
                filters=filters or {},
                lazy=True,
            )
        except (YesWeHackRawAPiError, YesWeHackApiClientError, requests.RequestException) as e:
            raise YesWeHackApiClientError(f'Unable to get reports for program {slug}') from e
        return self._get_detailed_reports(
            raw_reports=raw_reports,
        )

    def _get_detailed_reports(
        self,
        raw_reports: List[YesWeHackRawApiReport],
    ) -> List[Report]:
        return [
            self._get_detailed_report(
                report_id=raw_report.id,
            )
            for raw_report in raw_reports
        ]

    def _get_detailed_report(
        self,
        report_id: int,
    ) -> Report:
        try:
            raw_report = self._raw_client.get_report(
                report=report_id,
                lazy=True,
            )
        except (YesWeHackRawAPiError, requests.RequestException, TypeError) as e:
            raise YesWeHackApiClientError(f'Unable to get report #{report_id} details') from e
        return self._map_raw_report(
            raw_report=raw_report,
        )

    def _get_mapping_context(
        self,
    ) -> MappingContext:
        return MappingContext(
            yeswehack_domain=self._yeswehack_domain,
        )

    def _map_raw_report(
        self,
        raw_report: YesWeHackRawApiReport,
    ) -> Report:
        try:
            return map_raw_report(
                context=self._get_mapping_context(),
                raw_report=raw_report,
            )
        except TypeError as e:
            raise YesWeHackApiClientError(f'Unable to map report #{raw_report.id}') from e

    def _map_raw_attachment(
        self,
        raw_attachment: YesWeHackRawApiAttachment,
    ) -> Attachment:
        try:
            return map_raw_attachment(
                context=self._get_mapping_context(),
                raw_attachment=raw_attachment,
            )
        except TypeError as e:
            raise YesWeHackApiClientError(f'Unable to map attachment #{raw_attachment}') from e

    def _map_raw_logs(
        self,
        raw_logs: List[YesWeHackRawApiLog],
    ) -> List[Log]:
        try:
            return map_raw_logs(
                context=self._get_mapping_context(),
                raw_logs=raw_logs,
            )
        except TypeError as e:
            raise YesWeHackApiClientError(f'Unable to map logs #{raw_logs}') from e

    def put_report_tracking_status(
        self,
        report: Report,
        status: str,
        tracker_name: str,
        issue_id: str,
        issue_url: str,
        comment: str,
    ) -> None:
        """
        Update a report tracking status with information about a tracked issue.

        Args:
            report: a report
            status: a status (T, AFI, ...)
            tracker_name: a tracker name
            issue_id: an id of an issue from a tracker
            issue_url: an URL of an issue from a tracker
            comment: a comment

        Raises:
            YesWeHackApiClientError: if an error occurred when updating the status
        """
        self._ensure_login()
        try:
            response: requests.Response = report.raw_report.put_tracking_status(
                tracking_status=status,
                tracker_name=tracker_name,
                tracker_id=issue_id,
                tracker_url=issue_url,
                message=comment,
            )
        except (YesWeHackRawAPiError, requests.RequestException) as api_error:
            raise YesWeHackApiClientError(f'Unable to update report #{report.report_id} tracking status') from api_error
        try:
            data = response.json()
        except JSONDecodeError as decode_error:
            raise YesWeHackApiClientError(
                f'Unable to parse response after updating report #{report.report_id} tracking status',
            ) from decode_error
        if not isinstance(data, dict):
            raise YesWeHackApiClientError(
                (
                    f'Expecting {dict} from response to report #{report.report_id} tracking status update ; '
                    + f'got {type(data)}',
                ),
            )
        if not response.ok or 'errors' in data:
            message = data['message'] if 'message' in data and data['message'] else '[no error message]'
            raise YesWeHackApiClientError(
                f'Unable to update report #{report.report_id} tracking status: {message}',
            )

    def post_report_tracker_update(
        self,
        report: Report,
        tracker_name: str,
        issue_id: str,
        issue_url: str,
        token: str,
        comment: str,
    ) -> None:
        """
        Send a report tracker update with information about a tracked issue.

        Args:
            report: a report
            tracker_name: a tracker name
            issue_id: an id of an issue from a tracker
            issue_url: an URL of an issue from a tracker
            token: a token containing the state of the issue
            comment: a comment

        Raises:
            YesWeHackApiClientError: if an error occurred when sending the update
        """
        self._ensure_login()
        try:
            response = report.raw_report.post_tracker_update(
                tracker_name=tracker_name,
                tracker_id=issue_id,
                tracker_url=issue_url,
                token=token,
                message=comment,
            )
        except (YesWeHackRawAPiError, requests.RequestException) as api_error:
            raise YesWeHackApiClientError(f'Unable to send report #{report.report_id} tracker update') from api_error
        try:
            data = response.json()
        except JSONDecodeError as decode_error:
            raise YesWeHackApiClientError(
                f'Unable to parse response after report #{report.report_id} tracker update',
            ) from decode_error
        if not isinstance(data, dict):
            raise YesWeHackApiClientError(
                (
                    f'Expecting {dict} from response to report #{report.report_id} tracker update ; '
                    + f'got {type(data)}',
                ),
            )
        if not response.ok or 'errors' in data:
            message = data['message'] if 'message' in data and data['message'] else '[no error message]'
            raise YesWeHackApiClientError(
                f'Unable to update report #{report.report_id} tracker: {message}',
            )

    def post_report_attachment(
        self,
        report: Report,
        filename: str,
        file_content: bytes,
        file_type: Optional[str] = None,
    ) -> Attachment:
        """
        Send an attachment to a report.

        Args:
            report: a report
            filename: a file name
            file_content: file content
            file_type: an optional file type (mime)

        Returns:
            the attachment

        Raises:
            YesWeHackApiClientError: if an error occurred when sending the attachment
        """
        self._ensure_login()
        try:
            attachment = report.raw_report.post_attachment(
                filename=filename,
                file_content=file_content,
                file_type=file_type,
            )
        except (YesWeHackRawAPiError, requests.RequestException) as api_error:
            raise YesWeHackApiClientError(
                f'Unable to send attachment {filename} to report #{report.report_id}',
            ) from api_error
        return self._map_raw_attachment(
            raw_attachment=attachment,
        )

    def post_report_tracker_message(
        self,
        report: Report,
        tracker_name: str,
        issue_id: str,
        issue_url: str,
        comment: str,
        attachments: Optional[List[str]] = None,
    ) -> None:
        """
        Send a report tracker message with information about a tracked issue.

        Args:
            report: a report
            tracker_name: a tracker name
            issue_id: an id of an issue from a tracker
            issue_url: an URL of an issue from a tracker
            comment: a comment
            attachments: a list of attachment identifiers

        Raises:
            YesWeHackApiClientError: if an error occurred when sending the message
        """
        self._ensure_login()
        try:
            report.raw_report.post_tracker_message(
                tracker_name=tracker_name,
                tracker_id=issue_id,
                tracker_url=issue_url,
                message=comment,
                attachments=attachments,
            )
        except (YesWeHackRawAPiError, requests.RequestException) as api_error:
            raise YesWeHackApiClientError(f'Unable to send report #{report.report_id} tracker message') from api_error

    def put_status(
        self,
        report: Report,
        status: str,
        comment: str,
        attachments: Optional[List[str]] = None,
    ) -> List[Log]:
        """
        Update a report status.

        Args:
            report: a report
            status: a status
            comment: a comment
            attachments: a list of attachment identifiers

        Returns:
            a list of logs

        Raises:
            YesWeHackApiClientError: if an error occurred when updating the status
        """
        self._ensure_login()
        try:
            raw_logs = report.raw_report.put_status(
                status=status,
                message=comment,
                attachments=attachments,
            )
        except (YesWeHackRawAPiError, requests.RequestException) as api_error:
            raise YesWeHackApiClientError(f'Unable to update report #{report.report_id} status') from api_error
        return self._map_raw_logs(
            raw_logs=raw_logs,
        )



"""Models used for the configuration of a ServiceNow tracker."""
from typing import (
    Any,
    Optional,
    Text,
)

from ywh2bt.core.configuration.attribute import (
    Attribute,
    BoolAttributeType,
    StrAttributeType,
)
from ywh2bt.core.configuration.tracker import TrackerConfiguration
from ywh2bt.core.configuration.validator import (
    host_validator,
    not_blank_validator,
)


class ServiceNowConfiguration(TrackerConfiguration):
    """A class describing the configuration of a ServiceNow tracker."""

    host: StrAttributeType = Attribute.create(
        value_type=str,
        short_description='Instance host',
        description='Host of the ServiceNow instance',
        required=True,
        validator=host_validator,
    )
    login: StrAttributeType = Attribute.create(
        value_type=str,
        short_description='Login',
        description='User login for the ServiceNow instance',
        required=True,
        secret=False,
        validator=not_blank_validator,
    )
    password: StrAttributeType = Attribute.create(
        value_type=str,
        short_description='Password',
        description='User password for the ServiceNow instance',
        required=True,
        secret=True,
        validator=not_blank_validator,
    )
    use_ssl: BoolAttributeType = Attribute.create(
        value_type=bool,
        short_description='Use SSL',
        description='Use SSL connection',
        default=True,
    )
    verify: BoolAttributeType = Attribute.create(
        value_type=bool,
        short_description='Verify SSL',
        description='Verify SSL certs',
        default=True,
    )

    def __init__(
        self,
        host: Optional[Text] = None,
        login: Optional[Text] = None,
        password: Optional[Text] = None,
        use_ssl: Optional[bool] = None,
        verify: Optional[bool] = None,
        **kwargs: Any,
    ):
        """
        Initialize the configuration.

        Args:
            host: a ServiceNow instance host
            login: a ServiceNow user login
            password: a ServiceNow user password
            use_ssl: a flag indicating whether to use SSL/TLS connection
            verify: a flag indicating whether to check SSL/TLS connection
            kwargs: keyword arguments
        """
        super().__init__(**kwargs)
        self.host = host
        self.login = login
        self.password = password
        self.use_ssl = use_ssl
        self.verify = verify


TrackerConfiguration.register_subtype(
    subtype_name='servicenow',
    subtype_class=ServiceNowConfiguration,
)



"""Models and functions used for Json data SerDe."""
import json
from typing import Any, TextIO

from ywh2bt.core.serde import SerDe, SerDeError


class JsonSerDe(SerDe):
    """A Json Serializer/deserializer."""

    def serialize_to_stream(
        self,
        data: Any,
        stream: TextIO,
    ) -> None:
        """
        Write serialized data as Json into a stream.

        Args:
            data: data
            stream: a stream

        Raises:
            SerDeError: if an error occurred during writing or serialization
        """
        try:
            stream.write(json.dumps(data, indent=2))
        except Exception as e:
            raise SerDeError('JSON dump error') from e

    def deserialize_from_stream(
        self,
        stream: TextIO,
    ) -> Any:
        """
        Read and deserialize Json data from a stream.

        Args:
            stream: a stream

        Returns:
            the deserialized data

        Raises:
            SerDeError: if an error occurred during reading or deserialization
        """
        try:
            return json.load(stream)
        except Exception as e:
            raise SerDeError('JSON load error') from e



"""Models and functions used for attributes container dict entry GUI."""
from typing import Any, Optional

from PySide2.QtCore import Signal
from PySide2.QtWidgets import QFormLayout, QLabel, QLineEdit, QVBoxLayout, QWidget

from ywh2bt.core.configuration.attribute import AttributesContainer
from ywh2bt.gui.widgets.attribute.attributes_container_dict_entry import AttributesContainerDictEntry
from ywh2bt.gui.widgets.attribute.attributes_container_widget import AttributesContainerWidget
from ywh2bt.gui.widgets.typing import as_signal_instance


class AttributesContainerDictEntryWidget(QWidget):
    """Attributes container dict entry GUI."""

    key_changed: Signal = Signal(AttributesContainerDictEntry)
    value_changed: Signal = Signal(AttributesContainerDictEntry)

    _entry: AttributesContainerDictEntry
    _key_line_edit: QLineEdit
    _value_widget: AttributesContainerWidget

    def __init__(
        self,
        entry: AttributesContainerDictEntry,
        parent: Optional[QWidget] = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Initialize self.

        Args:
            entry: an entry
            parent: a parent widget
            args: extra arguments
            kwargs: extra keyword arguments
        """
        super().__init__(
            parent,
            *args,
            **kwargs,
        )
        self._entry = entry
        self._init_ui()

    def _init_ui(
        self,
    ) -> None:
        layout = QVBoxLayout(self)

        form_layout = QFormLayout()

        key_label = QLabel(
            'Key:',
            self,
        )
        self._key_line_edit = self._create_key_widget()
        self._key_line_edit.setText(self._entry.key)
        self._value_widget = self._create_value_widget()

        form_layout.addRow(key_label, self._key_line_edit)
        form_layout.addRow(self._value_widget)

        layout.addLayout(form_layout, 1)
        layout.addStretch(1)

    def _create_key_widget(
        self,
    ) -> QLineEdit:
        widget = QLineEdit(self)
        as_signal_instance(widget.textEdited).connect(
            self._on_key_changed,
        )
        return widget

    def _on_key_changed(
        self,
        value: str,
    ) -> None:
        self._entry.key = value
        as_signal_instance(self.key_changed).emit(
            self._entry,
        )

    def _create_value_widget(
        self,
    ) -> AttributesContainerWidget:
        widget = AttributesContainerWidget(
            parent=self,
            container_class=self._entry.value.__class__,
            container=self._entry.value,
        )
        as_signal_instance(widget.container_changed).connect(
            self._on_value_changed,
        )
        return widget

    def _on_value_changed(
        self,
        value: AttributesContainer,
    ) -> None:
        self._entry.value = value
        as_signal_instance(self.value_changed).emit(
            self._entry,
        )



"""Models and functions used for loading/writing configuration files."""
from dataclasses import dataclass
from typing import Dict, OrderedDict, Set, TextIO, Type

from ywh2bt.core.configuration.error import BaseAttributeError
from ywh2bt.core.configuration.root import RootConfiguration
from ywh2bt.core.exceptions import CoreException
from ywh2bt.core.loader import RootConfigurationLoader
from ywh2bt.core.schema.json import root_configuration_as_json_schema
from ywh2bt.core.schema.markdown import root_configuration_as_markdown
from ywh2bt.core.schema.protocol import SchemaDumpProtocol
from ywh2bt.core.schema.text import root_configuration_as_text
from ywh2bt.core.serde import SerDe
from ywh2bt.core.serializers.json import JsonSerDe
from ywh2bt.core.serializers.yaml import YamlSerDe


@dataclass
class AvailableFormat:
    """A class describing a core format and its supported file name extensions."""

    ser_de: Type[SerDe]
    extensions: Set[str]


AVAILABLE_FORMATS: Dict[str, AvailableFormat] = OrderedDict[str, AvailableFormat]({
    'yaml': AvailableFormat(
        ser_de=YamlSerDe,
        extensions={
            'yaml',
            'yml',
        },
    ),
    'json': AvailableFormat(
        ser_de=JsonSerDe,
        extensions={
            'json',
        },
    ),
})

AVAILABLE_SCHEMA_DUMP_FORMATS: Dict[str, SchemaDumpProtocol] = OrderedDict[str, SchemaDumpProtocol]({
    'text': root_configuration_as_text,
    'markdown': root_configuration_as_markdown,
    'json': root_configuration_as_json_schema,
})


def get_root_configuration_loader(
    file_format: str,
) -> RootConfigurationLoader:
    """
    Get a configuration loader for the given format.

    Args:
        file_format: a file format

    Returns:
        a configuration loader
    """
    serde = get_serde(
        file_format=file_format,
    )
    return RootConfigurationLoader(
        serde=serde,
    )


def get_serde(
    file_format: str,
) -> SerDe:
    """
    Get a serde for the given format.

    Args:
        file_format: a file format

    Returns:
        a serde

    Raises:
        CoreException: is the format is not supported
    """
    available_format = AVAILABLE_FORMATS.get(file_format)
    if available_format:
        return available_format.ser_de()
    raise CoreException(f'Unsupported file format {file_format}.')


def load_configuration(
    config_format: str,
    config_file: str,
) -> RootConfiguration:
    """
    Load a configuration file.

    Args:
        config_format: a file format
        config_file: a file

    Returns:
        a configuration

    Raises:
        CoreException: if the configuration couldn't be loaded
    """
    loader = get_root_configuration_loader(
        file_format=config_format,
    )
    try:
        with open(config_file, 'r') as f:
            return loader.load(
                stream=f,
            )
    except FileNotFoundError as e:
        raise CoreException('Configuration file not found') from e
    except BaseAttributeError as e:
        raise CoreException('Configuration invalid') from e


def write_message(
    stream: TextIO,
    message: str,
    end: str = '\n',
) -> None:
    """
    Write a message on the stream.

    Args:
        stream: a stream
        message: a message
        end: a string appended after the message, default a newline
    """
    stream.write(message)
    if end:
        stream.write(end)
    stream.flush()



"""Error related models and functions."""
import sys
import traceback
from io import StringIO
from typing import List, Optional, TextIO

from ywh2bt.core.configuration.error import (
    AttributesError,
    BaseAttributeError,
    InvalidAttributeError,
    MissingAttributeError,
    UnsupportedAttributeError,
)
from ywh2bt.core.configuration.subtypable import SubtypeError
from ywh2bt.core.core import write_message


def print_error_message(
    message: str,
    stream: Optional[TextIO] = None,
    end: str = '\n',
) -> None:
    """
    Print a message on STDERR.

    Args:
        message: a message
        stream: a stream to write the message to
        end: a string appended after the message, default a newline
    """
    if not stream:
        stream = sys.stderr
    write_message(
        stream=stream,
        message=message,
        end=end,
    )


def print_error(
    error: Exception,
    stream: Optional[TextIO] = None,
) -> None:
    """
    Print an error on STDERR.

    Args:
        error: an error
        stream: a stream to write the error to
    """
    if not stream:
        stream = sys.stderr
    print_error_message(
        stream=stream,
        message=_format_error(
            error=error,
        ),
    )


def error_to_string(
    error: Exception,
) -> str:
    """
    Get the full string representation of an error.

    Args:
        error: an error

    Returns:
        The string representation
    """
    error_stream = StringIO()
    print_error(
        stream=error_stream,
        error=error,
    )
    return error_stream.getvalue()


def _format_error(
    error: Exception,
) -> str:
    return '\n'.join(
        _get_formatted_error_items(
            error=error,
        ),
    )


def _get_formatted_error_items(
    error: Exception,
) -> List[str]:
    info = _extract_file_info(
        error=error,
    )
    class_name = error.__class__.__name__
    items = [f'{class_name}: {error} ({info})']
    if isinstance(error, BaseAttributeError):
        items.extend(_get_formatted_attribute_error_items(
            error=error,
        ))
    cause = getattr(error, '__cause__')  # noqa: B009
    if cause:
        cause_items = _get_formatted_error_items(
            error=cause,
        )
        items.extend(f'  {item}' for item in cause_items)
    return items


def _extract_file_info(
    error: Exception,
) -> str:
    if not hasattr(error, '__traceback__'):  # noqa: B009, WPS421
        return '?:?'
    tb = getattr(error, '__traceback__')  # noqa: B009
    stack_summary: traceback.StackSummary = traceback.extract_tb(tb)
    if stack_summary:
        frame_summary: traceback.FrameSummary = stack_summary[0]
        return f'{frame_summary.filename}:{frame_summary.lineno}'
    return '?:?'


def _get_formatted_attribute_error_items(
    error: BaseAttributeError,
) -> List[str]:
    if isinstance(error, AttributesError):
        return _get_formatted_attributes_error_items(
            error=error,
        )
    elif isinstance(error, (InvalidAttributeError, MissingAttributeError, SubtypeError, UnsupportedAttributeError)):
        return []
    return [
        f'Nonformatted error {error.__class__.__name__}: {error}',
    ]


def _get_formatted_attributes_error_items(
    error: AttributesError,
) -> List[str]:
    items = []
    for error_key, error_item in error.errors.items():
        formatted_item = _get_formatted_attribute_error_items(
            error=error_item,
        )
        items.append(f'  - {error_key}: {error_item}')
        items.extend(f'    {item}' for item in formatted_item)
    return items



"""Models and functions used for decryption."""
from ywh2bt.core.crypt.error import CryptError
from ywh2bt.core.crypt.key import Key


class Decryptor:
    """A simple decryptor."""

    _key: Key

    def __init__(
        self,
        key: Key,
    ):
        """
        Initialize self.

        Args:
            key: a cryptographic key

        Raises:
            CryptError: if an error occurred
        """
        if not isinstance(key, Key):
            raise CryptError(f'Key should be {Key}. {type(key)} given.')
        self._key = key

    def decrypt(
        self,
        encrypted_bytes: bytes,
    ) -> bytes:
        """
        Decrypt a message.

        Args:
            encrypted_bytes: an encrypted message

        Raises:
            CryptError: if an error occurred

        Returns:
            The decrypted message.
        """
        if not isinstance(encrypted_bytes, bytes):
            raise CryptError(f'Encrypted should be {bytes}. {type(encrypted_bytes)} given.')
        key_generator = self._key.generator()
        encrypted_size = len(encrypted_bytes)
        return bytes([
            next(key_generator) ^ encrypted_bytes[i]
            for i in range(encrypted_size)
        ])



"""Models and functions used for states decryption."""
import base64
import json
import re
from typing import Optional, Type, TypeVar

from ywh2bt.core.crypt.decrypt import Decryptor
from ywh2bt.core.crypt.error import CryptError
from ywh2bt.core.crypt.key import Key
from ywh2bt.core.state.error import StateError
from ywh2bt.core.state.state import State

StateType = TypeVar('StateType', bound=State, covariant=True)


class StateDecryptor:
    """A simple state decryptor."""

    _encrypted_re = re.compile(
        r'\[YWH2BT:S:(.+)]',
    )

    @classmethod
    def decrypt(
        cls,
        encrypted_state: str,
        key: str,
        state_type: Type[StateType],
    ) -> Optional[StateType]:
        """
        Decrypt a state with a key.

        Args:
            encrypted_state: an encrypted state
            key: a cryptographic key
            state_type: a type of the decrypted state

        Raises:
            StateError: if the state could not be decrypted

        Returns:
            The decrypted state.
        """
        encrypted_base64 = cls._extract_encrypted(
            encrypted_state=encrypted_state,
        )
        if not encrypted_base64:
            return None
        encrypted = base64.b64decode(encrypted_base64)
        decryptor = Decryptor(
            key=Key.from_str(
                key=key,
            ),
        )
        try:
            decrypted_data = decryptor.decrypt(encrypted)
        except CryptError as decrypt_error:
            raise StateError(f'Unable to decrypt state {encrypted_state}') from decrypt_error
        try:
            class_name, data = json.loads(decrypted_data)
        except (json.JSONDecodeError, TypeError) as json_error:
            raise StateError(f'Unable to deserialize state {encrypted_state}') from json_error
        if class_name != state_type.__name__:
            raise StateError(f'Invalid state type {class_name}')
        if not isinstance(data, dict):
            raise StateError('Invalid state data')
        return state_type(**data)  # type: ignore

    @classmethod
    def _extract_encrypted(
        cls,
        encrypted_state: str,
    ) -> Optional[str]:
        match = cls._encrypted_re.search(encrypted_state)
        if match:
            return match.group(1)
        return None



"""Models used for the configuration of a GitLab tracker."""
from typing import Any, Optional, Text

from ywh2bt.core.configuration.attribute import Attribute, BoolAttributeType, StrAttributeType
from ywh2bt.core.configuration.tracker import TrackerConfiguration
from ywh2bt.core.configuration.validator import not_blank_validator, url_validator


class GitLabConfiguration(TrackerConfiguration):
    """A class describing the configuration of a GitLab tracker."""

    url: StrAttributeType = Attribute.create(
        value_type=str,
        short_description='API URL',
        description='Base URL of the GitLab server',
        default='https://gitlab.com',
        validator=url_validator,
    )
    token: StrAttributeType = Attribute.create(
        value_type=str,
        short_description='API token',
        description='User private token for the GitLab API',
        required=True,
        secret=True,
        validator=not_blank_validator,
    )
    project: StrAttributeType = Attribute.create(
        value_type=str,
        short_description='Project path',
        description='Path to the project on GitLab',
        required=True,
        validator=not_blank_validator,
    )
    verify: BoolAttributeType = Attribute.create(
        value_type=bool,
        short_description='Verify TLS',
        description="Verify server's TLS certificate",
        default=True,
    )
    confidential: BoolAttributeType = Attribute.create(
        value_type=bool,
        short_description='Confidential issues',
        description='Mark created issues as confidential',
        default=False,
    )

    def __init__(
        self,
        url: Optional[Text] = None,
        token: Optional[Text] = None,
        project: Optional[Text] = None,
        verify: Optional[bool] = None,
        confidential: Optional[bool] = None,
        **kwargs: Any,
    ):
        """
        Initialize the configuration.

        Args:
            url: a GitLab API URL
            token: a GitLab API token
            project: a GitLab project name
            verify: a flag indicating whether to check SSL/TLS connection
            confidential: a flag indicating whether to mark created issues as confidential
            kwargs: keyword arguments
        """
        super().__init__(**kwargs)
        self.url = url
        self.token = token
        self.project = project
        self.verify = verify
        self.confidential = confidential


TrackerConfiguration.register_subtype(
    subtype_name='gitlab',
    subtype_class=GitLabConfiguration,
)



"""Models and functions used for formatting reports data into markdown."""

from html import escape as html_escape
from string import Template

from ywh2bt.core.api.formatter.formatter import ReportMessageFormatter
from ywh2bt.core.html import ywh_html_to_markdown

REPORT_TITLE_TEMPLATE = '${local_id} : ${title}'
REPORT_DESCRIPTION_TEMPLATE = """
| Title | ${local_id} : ${title} |
|-------|---------------------|
| Priority | ${priority_name} |
| ${bug_type_label} | [${bug_type_name}](${bug_type_link}) &#8594; [Remediation](${bug_type_remediation_link}) |
| ${scope_label} | ${scope} |
| Severity | ${cvss_criticity}, score: ${cvss_score}, vector: ${cvss_vector} |
| ${end_point_label} | ${end_point} |
| ${vulnerable_part_label} | ${vulnerable_part} |
| ${part_name_label} | ${part_name} |
| ${payload_sample_label} | ${payload_sample} |
| Technical Environment | ${technical_environment} |

${description}
"""
CLOSE_LOG_TEMPLATE = """
**Report closed**:

${old_status} -> ${new_status}

**Comment**:

${comment}
"""
COMMENT_BODY_TEMPLATE = """
**Date**: ${date}

${body}
"""
COMMENT_LOG_TEMPLATE = """
**Comment**:

${comment}
"""
CVSS_UPDATE_LOG_TEMPLATE = """
**CVSS updated**:

| Detail | Old value | New value |
|--------|-----------|-----------|
| **Severity** | ${old_cvss_criticity} | ${new_cvss_criticity} |
| **Score** | ${old_cvss_score} | ${new_cvss_score} |
"""
STATUS_UPDATE_LOG_TEMPLATE = """
**Status updated**:

${old_status} -> ${new_status}

**Comment**:

${comment}
"""
FIX_VERIFIED_LOG_TEMPLATE = """
**Fix verification**:

${verified}

**Comment**:

${comment}
"""
DETAILS_UPDATE_LOG_TEMPLATE = """
**Details updated**:

| Detail | Old value | New value |
|--------|-----------|-----------|
${details_lines}
"""
DETAILS_UPDATE_LOG_LINE_TEMPLATE = """| **${updated_property}** | ${old_value} | ${new_value} |
"""
PRIORITY_UPDATE_LOG_TEMPLATE = """
**Priority updated**:

${new_priority}
"""
REWARD_LOG_TEMPLATE = """
**Comment from reward**:

${comment}
"""


def _html_transformer(
    value: str,
) -> str:
    return html_escape(value)


class ReportMessageMarkdownFormatter(ReportMessageFormatter):
    """A report formatter to markdown."""

    def __init__(self) -> None:
        """Initialize self."""
        super().__init__(
            report_title_template=Template(REPORT_TITLE_TEMPLATE),
            report_description_template=Template(REPORT_DESCRIPTION_TEMPLATE),
            close_log_template=Template(CLOSE_LOG_TEMPLATE),
            comment_body_template=Template(COMMENT_BODY_TEMPLATE),
            comment_log_template=Template(COMMENT_LOG_TEMPLATE),
            cvss_update_log_template=Template(CVSS_UPDATE_LOG_TEMPLATE),
            status_update_log_template=Template(STATUS_UPDATE_LOG_TEMPLATE),
            fix_verified_log_template=Template(FIX_VERIFIED_LOG_TEMPLATE),
            details_update_log_template=Template(DETAILS_UPDATE_LOG_TEMPLATE),
            details_update_log_line_template=Template(DETAILS_UPDATE_LOG_LINE_TEMPLATE),
            priority_update_log_template=Template(PRIORITY_UPDATE_LOG_TEMPLATE),
            reward_log_template=Template(REWARD_LOG_TEMPLATE),
            value_transformer=_html_transformer,
        )

    def transform_report_description_html(
        self,
        description_html: str,
    ) -> str:
        """
        Transform the report description.

        Args:
            description_html: a report description in HTML

        Returns:
            a transformed description
        """
        return ywh_html_to_markdown(
            html=description_html,
        )

    def transform_html(
        self,
        html: str,
    ) -> str:
        """
        Transform from HTML.

        Args:
            html: HTML

        Returns:
            HTML transformed to markdown
        """
        return ywh_html_to_markdown(
            html=html,
        )
